# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

# This file contains several 2.x/3.x compatibility checkstyle violations for a reason
# checkstyle: noqa

from __future__ import absolute_import

import os
import re
import sys
from abc import ABCMeta
from sys import version_info as sys_version_info

from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import (
        IO,
        AnyStr,
        BinaryIO,
        Callable,
        Deque,
        List,
        Optional,
        Sequence,
        Text,
        Tuple,
        Type,
    )


try:
    # Python 2.x
    from ConfigParser import ConfigParser as ConfigParser
except ImportError:
    # Python 3.x
    from configparser import ConfigParser as ConfigParser  # type: ignore[import, no-redef]


AbstractClass = ABCMeta("AbstractClass", (object,), {})
PY2 = sys_version_info[0] == 2
PY3 = sys_version_info[0] == 3

string = cast("Tuple[Type, ...]", (str,) if PY3 else (str, unicode))  # type: ignore[name-defined]
text = cast("Type[Text]", str if PY3 else unicode)  # type: ignore[name-defined]


if PY2:
    from collections import Iterable as Iterable
    from collections import MutableSet as MutableSet
    from collections import deque
else:
    from collections.abc import Iterable as Iterable
    from collections.abc import MutableSet as MutableSet

if PY2:

    def to_bytes(st, encoding="utf-8"):
        # type: (AnyStr, Text) -> bytes
        if isinstance(st, unicode):
            return st.encode(encoding)
        elif isinstance(st, bytes):
            return st
        else:
            raise ValueError("Cannot convert %s to bytes" % type(st))

    def to_unicode(st, encoding="utf-8"):
        # type: (AnyStr, Text) -> Text
        if isinstance(st, unicode):
            return st
        elif isinstance(st, (str, bytes)):
            return unicode(st, encoding)
        else:
            raise ValueError("Cannot convert %s to a unicode string" % type(st))

else:

    def to_bytes(st, encoding="utf-8"):
        # type: (AnyStr, Text) -> bytes
        if isinstance(st, str):
            return st.encode(encoding)
        elif isinstance(st, bytes):
            return st
        else:
            raise ValueError("Cannot convert %s to bytes." % type(st))

    def to_unicode(st, encoding="utf-8"):
        # type: (AnyStr, Text) -> Text
        if isinstance(st, str):
            return st
        elif isinstance(st, bytes):
            return str(st, encoding)
        else:
            raise ValueError("Cannot convert %s to a unicode string" % type(st))


_PY3_EXEC_FUNCTION = """
def exec_function(ast, globals_map):
  locals_map = globals_map
  exec ast in globals_map, locals_map
  return locals_map
"""

if PY3:

    def exec_function(ast, globals_map):
        locals_map = globals_map
        exec (ast, globals_map, locals_map)
        return locals_map

else:

    def exec_function(ast, globals_map):
        raise AssertionError("Expected this function to be re-defined at runtime.")

    # This will result in `exec_function` being re-defined at runtime.
    eval(compile(_PY3_EXEC_FUNCTION, "<exec_function>", "exec"))


if PY3:
    from urllib import parse as urlparse
    from urllib.error import HTTPError as HTTPError
    from urllib.parse import quote as _url_quote
    from urllib.parse import unquote as _url_unquote
    from urllib.request import FileHandler as FileHandler
    from urllib.request import HTTPBasicAuthHandler as HTTPBasicAuthHandler
    from urllib.request import HTTPDigestAuthHandler as HTTPDigestAuthHandler
    from urllib.request import HTTPPasswordMgrWithDefaultRealm as HTTPPasswordMgrWithDefaultRealm
    from urllib.request import HTTPSHandler as HTTPSHandler
    from urllib.request import ProxyHandler as ProxyHandler
    from urllib.request import Request as Request
    from urllib.request import build_opener as build_opener
else:
    from urllib import quote as _url_quote
    from urllib import unquote as _url_unquote

    import urlparse as urlparse
    from urllib2 import FileHandler as FileHandler
    from urllib2 import HTTPBasicAuthHandler as HTTPBasicAuthHandler
    from urllib2 import HTTPDigestAuthHandler as HTTPDigestAuthHandler
    from urllib2 import HTTPError as HTTPError
    from urllib2 import HTTPPasswordMgrWithDefaultRealm as HTTPPasswordMgrWithDefaultRealm
    from urllib2 import HTTPSHandler as HTTPSHandler
    from urllib2 import ProxyHandler as ProxyHandler
    from urllib2 import Request as Request
    from urllib2 import build_opener as build_opener

url_unquote = _url_unquote
url_quote = _url_quote

if PY3:
    from queue import Queue as Queue

    # The `os.sched_getaffinity` function appears to be supported on Linux but not OSX.
    if not hasattr(os, "sched_getaffinity"):
        from os import cpu_count as cpu_count
    else:

        def cpu_count():
            # type: () -> Optional[int]
            # The set of CPUs accessible to the current process (pid 0).
            cpu_set = os.sched_getaffinity(0)
            return len(cpu_set)

else:
    from multiprocessing import cpu_count as cpu_count

    from Queue import Queue as Queue

WINDOWS = os.name == "nt"


# Universal newlines is the default in Python 3.
MODE_READ_UNIVERSAL_NEWLINES = "rU" if PY2 else "r"


def _get_stdio_bytes_buffer(stdio):
    # type: (IO[str]) -> BinaryIO
    return cast("BinaryIO", getattr(stdio, "buffer", stdio))


def get_stdout_bytes_buffer():
    # type: () -> BinaryIO
    return _get_stdio_bytes_buffer(sys.stdout)


def get_stderr_bytes_buffer():
    # type: () -> BinaryIO
    return _get_stdio_bytes_buffer(sys.stderr)


if PY3:
    is_valid_python_identifier = str.isidentifier
else:

    def is_valid_python_identifier(text):
        # type: (str) -> bool

        # N.B.: Python 2.7 only supports ASCII characters so the check is easy and this is probably
        # why it's nt in the stdlib.
        # See: https://docs.python.org/2.7/reference/lexical_analysis.html#identifiers
        return re.match(r"^[_a-zA-Z][_a-zA-Z0-9]*$", text) is not None


if PY2:

    def indent(
        text,  # type: Text
        prefix,  # type: Text
        predicate=None,  # type: Optional[Callable[[Text], bool]]
    ):
        add_prefix = predicate if predicate else lambda line: bool(line.strip())
        return "".join(
            prefix + line if add_prefix(line) else line for line in text.splitlines(True)
        )

else:
    from textwrap import indent as indent


if PY3:
    from os.path import commonpath as commonpath
else:

    def commonpath(paths):
        # type: (Sequence[Text]) -> Text
        if not paths:
            raise ValueError("The paths given must be a non-empty sequence")
        if len(paths) == 1:
            return paths[0]
        if len({os.path.isabs(path) for path in paths}) > 1:
            raise ValueError(
                "Can't mix absolute and relative paths, given:\n{paths}".format(
                    paths="\n".join(paths)
                )
            )

        def components(path):
            # type: (Text) -> Iterable[Text]

            pieces = deque()  # type: Deque[Text]

            def append(piece):
                if piece and piece != ".":
                    pieces.appendleft(piece)

            head, tail = os.path.split(path)
            append(tail)
            while head:
                if "/" == head:
                    append(head)
                    break
                head, tail = os.path.split(head)
                append(tail)
            return pieces

        prefix = []  # type: List[Text]
        for atoms in zip(*(components(path) for path in paths)):
            if len(set(atoms)) == 1:
                prefix.append(atoms[0])
            else:
                break
        if not prefix:
            return ""
        return os.path.join(*prefix)


if PY3:
    from shlex import quote as _shlex_quote
else:
    from pipes import quote as _shlex_quote

shlex_quote = _shlex_quote



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import re

from colors import green

from pex.cli.testing import run_pex3
from pex.resolve.lockfile import json_codec
from pex.testing import run_pex_command


def test_preserve_pip_download_log():
    # type: () -> None

    result = run_pex3("lock", "create", "ansicolors==1.1.8", "--preserve-pip-download-log")
    result.assert_success()

    match = re.search(
        r"^pex: Preserving `pip download` log at (?P<log_path>.*)$", result.error, re.MULTILINE
    )
    assert match is not None
    log_path = match.group("log_path")
    assert os.path.exists(log_path)
    expected_url = (
        "https://files.pythonhosted.org/packages/53/18/"
        "a56e2fe47b259bb52201093a3a9d4a32014f9d85071ad07e9d60600890ca/"
        "ansicolors-1.1.8-py2.py3-none-any.whl"
    )
    expected_algorithm = "sha256"
    expected_hash = "00d2dde5a675579325902536738dd27e4fac1fd68f773fe36c21044eb559e187"
    with open(log_path) as fp:
        # N.B.: Modern Pip excludes hashes from logged URLs when the index serves up PEP-691 json
        # responses.
        assert re.search(
            r"Added ansicolors==1\.1\.8 from {url}(?:#{algorithm}={hash})? to build tracker".format(
                url=re.escape(expected_url),
                algorithm=re.escape(expected_algorithm),
                hash=re.escape(expected_hash),
            ),
            fp.read(),
        )

    lockfile = json_codec.loads(result.output)
    assert 1 == len(lockfile.locked_resolves)

    locked_resolve = lockfile.locked_resolves[0]
    assert 1 == len(locked_resolve.locked_requirements)

    locked_requirement = locked_resolve.locked_requirements[0]
    artifacts = tuple(locked_requirement.iter_artifacts())
    assert 1 == len(artifacts)

    artifact = artifacts[0]
    assert expected_url == artifact.url
    assert expected_algorithm == artifact.fingerprint.algorithm
    assert expected_hash == artifact.fingerprint.hash


def test_preserve_pip_download_log_none():
    # type: () -> None

    result = run_pex_command(
        args=[
            "ansicolors==1.1.8",
            "--preserve-pip-download-log",
            "--",
            "-c",
            "import colors; print(colors.green('42'))",
        ],
        quiet=True,
    )
    result.assert_success()
    assert green("42") == result.output.strip()
    assert (
        "pex: The `pip download` log is not being utilized, to see more `pip download` details, "
        "re-run with more Pex verbosity (more `-v`s).\n"
    ) in result.error



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
import sys

from pex.common import touch
from pex.typing import TYPE_CHECKING
from pex.vendor import VendorSpec

if TYPE_CHECKING:
    from typing import Any


def test_pinned():
    # type: () -> None
    vendor_spec = VendorSpec.pinned("foo", "1.2.3")
    assert "foo" == vendor_spec.key
    assert "foo==1.2.3" == vendor_spec.requirement


def test_git():
    # type: () -> None
    vendor_spec = VendorSpec.git(
        repo="https://github.com/foo.git", commit="da39a3ee", project_name="bar"
    )
    assert "bar" == vendor_spec.key
    assert "git+https://github.com/foo.git@da39a3ee#egg=bar" == vendor_spec.requirement
    assert "git+https://github.com/foo.git@da39a3ee#egg=bar" == vendor_spec.prepare()


def test_git_prep_command(tmpdir):
    # type: (Any) -> None
    repo = os.path.join(str(tmpdir), "repo")
    subprocess.check_call(["git", "init", repo])
    assert os.path.isdir(repo)

    subprocess.check_call(["git", "config", "user.email", "you@example.com"], cwd=repo)
    subprocess.check_call(["git", "config", "user.name", "Your Name"], cwd=repo)

    touch(os.path.join(repo, "README"))
    subprocess.check_call(["git", "add", "README"], cwd=repo)
    subprocess.check_call(["git", "commit", "-m", "Initial Commit."], cwd=repo)
    commit = subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=repo).decode("utf-8").strip()

    prep_file = os.path.join(repo, "prep")
    assert not os.path.exists(prep_file)

    vendor_spec = VendorSpec.git(
        repo=repo,
        commit=commit,
        project_name="bar",
        prep_command=[sys.executable, "-c", "fp = open('prep', 'w'); fp.close()"],
    )
    assert not os.path.exists(prep_file)

    assert "git+{repo}@{commit}#egg=bar".format(repo=repo, commit=commit) == vendor_spec.requirement
    assert not os.path.exists(prep_file)

    clone = vendor_spec.prepare()
    assert (
        commit
        == subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=clone).decode("utf-8").strip()
    )
    assert not os.path.exists(prep_file)
    assert os.path.isfile(os.path.join(clone, "prep"))



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, division, print_function, unicode_literals

import subprocess
import sys
from typing import Any

from pex.testing import IntegResults


def run_pex3(
    *args,  # type: str
    **popen_kwargs  # type: Any
):
    # type: (...) -> IntegResults
    process = subprocess.Popen(
        args=[sys.executable, "-mpex.cli"] + list(args),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **popen_kwargs
    )
    stdout, stderr = process.communicate()
    return IntegResults(
        output=stdout.decode("utf-8"), error=stderr.decode("utf-8"), return_code=process.returncode
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import sys
import traceback

from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Any, Callable, Text, TypeVar, Union

    import attr  # vendor:skip

    _T = TypeVar("_T")
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Result(object):
    exit_code = attr.ib()  # type: int
    _message = attr.ib(default="")  # type: Text

    @property
    def is_error(self):
        # type: () -> bool
        return self.exit_code != 0

    def maybe_display(self):
        # type: () -> None
        if not self._message:
            return
        print(self._message, file=sys.stderr if self.is_error else sys.stdout)

    def __str__(self):
        # type: () -> str
        return str(self._message)


class Ok(Result):
    def __init__(self, message=""):
        # type: (Text) -> None
        super(Ok, self).__init__(exit_code=0, message=message)


class Error(Result):
    def __init__(
        self,
        message="",  # type: Text
        exit_code=1,  # type: int
    ):
        # type: (...) -> None
        if exit_code == 0:
            raise ValueError("An Error must have a non-zero exit code; given: {}".format(exit_code))
        super(Error, self).__init__(exit_code=exit_code, message=message)


# Frozen exception types don't work under 3.11.0b1 which sets the `__traceback__` attribute after
# construction in some cases.
@attr.s
class ResultError(Exception):
    """Wraps an Error in an exception for use in short-circuiting via `try_` / `catch`."""

    error = attr.ib()  # type: Error

    def __str__(self):
        # type: () -> str
        return str(self.error)


def try_(result):
    # type: (Union[_T, Error]) -> _T
    """Return the result unless it's an `Error`.

    Paired with `catch` at a higher layer, this allows for ~checked exceptions without the line
    noise of try / except handling.

    :param result: a result containing a successful value or else an error.
    :return: the result if not an error.
    :raise: :class:`ResultError` if the result is an error.
    """
    if isinstance(result, Error):
        raise ResultError(error=result)
    return result


def catch(
    func,  # type: Callable[..., _T]
    *args,  # type: Any
    **kwargs  # type: Any
):
    # type: (...) -> Union[_T, Error]
    """Execute the given function in a context that handles short-circuit error returns.

    :param func: The function to execute.
    :param args: Any arguments to be passed to func.
    :param kwargs: Any keyword arguments to be passed to func.
    :return: The result of the function or an error if any error was returned in the call stack of
             the function.
    """
    try:
        return func(*args, **kwargs)
    except ResultError as e:
        return e.error
    except Exception as e:
        if ENV.PEX_VERBOSE > 0:
            traceback.print_exc()
        return Error(str(e))



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
#
# OrderedSet recipe referenced in the Python standard library docs (bottom):
#     http://docs.python.org/library/collections.html
#
# Copied from recipe code found here: http://code.activestate.com/recipes/576694/ with small
# modifications
#

from __future__ import absolute_import

from collections import OrderedDict

from pex.compatibility import MutableSet
from pex.typing import TYPE_CHECKING, Generic, cast

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, Optional


# N.B.: The item type is mixed in only at type checking time below due to class hierarchy metaclass
# conflicts. As such, we use Any here.
class _OrderedSet(MutableSet):
    def __init__(self, iterable=None):
        # type: (Optional[Iterable]) -> None
        self._data = OrderedDict()  # type: OrderedDict[Any, None]
        if iterable is not None:
            self.update(iterable)

    def __len__(self):
        # type: () -> int
        return len(self._data)

    def __contains__(self, key):
        # type: (Any) -> bool
        return key in self._data

    def add(self, key):
        # type: (Any) -> None
        self._data[key] = None

    def update(self, iterable):
        # type: (Iterable[Any]) -> None
        for key in iterable:
            self.add(key)

    def discard(self, key):
        # type: (Any) -> None
        self._data.pop(key, None)

    def __iter__(self):
        # type: () -> Iterator[Any]
        return iter(self._data)

    def __reversed__(self):
        # type: () -> Iterator[Any]
        return reversed(self._data)

    def pop(self, last=True):
        # type: (bool) -> Any
        if not self:
            raise KeyError("set is empty")
        key, _ = self._data.popitem(last=last)
        return key

    def __repr__(self):
        # type: () -> str
        if not self:
            return "{}()".format(
                self.__class__.__name__,
            )
        return "{}({!r})".format(self.__class__.__name__, list(self))

    def __eq__(self, other):
        # type: (Any) -> bool
        if type(other) != type(self):
            return NotImplemented
        # TODO(John Sirois): Type __eq__ as returning Union[bool, NotImplemented] when MyPy fixes:
        #  https://github.com/python/mypy/issues/4791
        return self._data == other._data  # type: ignore[no-any-return]


if TYPE_CHECKING:
    from typing import TypeVar

    _I = TypeVar("_I")

    # N.B.: We mix in Generic only in type checking mode since it uses a custom metaclass in
    # production which is inconsistent with _OrderedSet's base of MutableSet which has an ABCMeta
    # metaclass.
    class OrderedSet(_OrderedSet, Generic["_I"]):
        def __init__(self, iterable=None):
            # type: (Optional[Iterable[_I]]) -> None
            super(OrderedSet, self).__init__(iterable=iterable)

        def add(self, key):
            # type: (_I) -> None
            super(OrderedSet, self).add(key)

        def update(self, iterable):
            # type: (Iterable[_I]) -> None
            super(OrderedSet, self).update(iterable)

        def discard(self, key):
            # type: (_I) -> None
            super(OrderedSet, self).discard(key)

        def __iter__(self):
            # type: () -> Iterator[_I]
            return super(OrderedSet, self).__iter__()

        def __reversed__(self):
            # type: () -> Iterator[_I]
            return super(OrderedSet, self).__reversed__()

        def pop(self, last=True):
            # type: (bool) -> _I
            return cast("_I", super(OrderedSet, self).pop(last=last))

else:
    OrderedSet = _OrderedSet



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex import sh_boot
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraints
from pex.orderedset import OrderedSet
from pex.pep_425 import CompatibilityTags
from pex.pep_508 import MarkerEnvironment
from pex.platforms import Platform
from pex.sh_boot import PythonBinaryName
from pex.targets import CompletePlatform, Targets
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, List


def calculate_binary_names(
    targets=Targets(),  # type: Targets
    interpreter_constraints=(),  # type: Iterable[str]
):
    return list(
        sh_boot._calculate_applicable_binary_names(
            targets=targets,
            interpreter_constraints=InterpreterConstraints.parse(*interpreter_constraints),
        )
    )


def expected(*names):
    # type: (*PythonBinaryName) -> List[str]

    current_interpreter_identity = PythonInterpreter.get().identity

    # The default expected set of interpreters should always include:
    # 1. The targeted interpreters or the current interpreter if none were explicitly targeted.
    # 2. The interpreters Pex supports, CPython 1st since its faster to boot and re-exec with even
    #    if PyPy is what will be re-exec'd to. Also, newest version 1st since these are more likely
    #    to match the final interpreter than the reverse.
    # 3. Partially abbreviated names for the above in corresponding order.
    # 4. Fully abbreviated names for the above in corresponding order.
    all_names = OrderedSet(name.render(version_components=2) for name in names)
    if not names:
        all_names.add(current_interpreter_identity.binary_name(version_components=2))
    all_names.update(
        [
            "python3.11",
            "python3.10",
            "python3.9",
            "python3.8",
            "python3.7",
            "python3.6",
            "python3.5",
            "python2.7",
            "pypy3.11",
            "pypy3.10",
            "pypy3.9",
            "pypy3.8",
            "pypy3.7",
            "pypy3.6",
            "pypy3.5",
            "pypy2.7",
        ]
    )

    if names:
        all_names.update(name.render(version_components=1) for name in names)
    else:
        all_names.add(current_interpreter_identity.binary_name(version_components=1))
    all_names.update(
        [
            "python3",
            "python2",
            "pypy3",
            "pypy2",
        ]
    )

    if names:
        all_names.update(name.render(version_components=0) for name in names)
    else:
        all_names.add(current_interpreter_identity.binary_name(version_components=0))
    all_names.update(
        [
            "python",
            "pypy",
        ]
    )

    return list(all_names)


def test_calculate_no_targets_no_ics():
    # type: () -> None

    assert expected() == calculate_binary_names()


def test_calculate_platforms_no_ics():
    # type: () -> None

    assert expected(
        PythonBinaryName(name="python", version=(3, 6)),
        PythonBinaryName(name="pypy", version=(2, 7)),
    ) == calculate_binary_names(
        Targets(
            platforms=(
                Platform.create("macosx-10.13-x86_64-cp-36-cp36m"),
                Platform.create("linux-x86_64-pp-27-pypy_73"),
            )
        )
    )


def test_calculate_interpreters_no_ics(
    py27,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
    py38,  # type: PythonInterpreter
):
    # type: (...) -> None

    assert (
        expected(
            PythonBinaryName(name="python", version=(2, 7)),
            PythonBinaryName(name="python", version=(3, 10)),
            PythonBinaryName(name="python", version=(3, 8)),
        )
        == calculate_binary_names(targets=Targets(interpreters=(py27, py310, py38)))
    )


def test_calculate_no_targets_ics():
    # type: () -> None

    assert (
        expected(
            PythonBinaryName(name="python", version=(3, 7)),
            PythonBinaryName(name="pypy", version=(3, 7)),
            PythonBinaryName(name="python", version=(3, 8)),
            PythonBinaryName(name="pypy", version=(3, 8)),
            PythonBinaryName(name="python", version=(3, 9)),
            PythonBinaryName(name="pypy", version=(3, 9)),
            PythonBinaryName(name="pypy", version=(3, 6)),
        )
        == calculate_binary_names(interpreter_constraints=[">=3.7,<3.10", "PyPy==3.6.*"])
    )


def test_calculate_mixed(py27):
    # type: (PythonInterpreter) -> None

    assert expected(
        PythonBinaryName(name="python", version=(2, 7)),
        PythonBinaryName(name="pypy", version=(3, 8)),
        PythonBinaryName(name="python", version=(3, 6)),
        PythonBinaryName(name="pypy", version=(3, 7)),
    ) == calculate_binary_names(
        targets=Targets(
            interpreters=(py27,),
            complete_platforms=(
                CompletePlatform.create(
                    marker_environment=MarkerEnvironment(
                        platform_python_implementation="PyPy", python_version="3.8"
                    ),
                    # N.B.: Unused by the code under test, but we need to supply at least 1 tag.
                    supported_tags=CompatibilityTags.from_strings(("py3-none-any",)),
                ),
            ),
        ),
        interpreter_constraints=["CPython==3.6.*", "PyPy>=3.7,<3.9"],
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os

from pex.dist_metadata import Requirement
from pex.orderedset import OrderedSet
from pex.pip.version import PipVersionValue
from pex.requirements import LocalProjectRequirement
from pex.resolve.locked_resolve import LocalProjectArtifact, LockedResolve, LockStyle, TargetSystem
from pex.resolve.resolved_requirement import Pin
from pex.resolve.resolver_configuration import ResolverVersion
from pex.sorted_tuple import SortedTuple
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, Iterable, Mapping, Optional, Union

    import attr  # vendor:skip

    from pex.requirements import ParsedRequirement
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Lockfile(object):
    @classmethod
    def create(
        cls,
        pex_version,  # type: str
        style,  # type: LockStyle.Value
        requires_python,  # type: Iterable[str]
        target_systems,  # type: Iterable[TargetSystem.Value]
        pip_version,  # type: PipVersionValue
        resolver_version,  # type: ResolverVersion.Value
        requirements,  # type: Iterable[Union[Requirement, ParsedRequirement]]
        constraints,  # type: Iterable[Requirement]
        allow_prereleases,  # type: bool
        allow_wheels,  # type: bool
        allow_builds,  # type: bool
        prefer_older_binary,  # type: bool
        use_pep517,  # type: Optional[bool]
        build_isolation,  # type: bool
        transitive,  # type: bool
        locked_resolves,  # type: Iterable[LockedResolve]
        source=None,  # type: Optional[str]
    ):
        # type: (...) -> Lockfile

        pin_by_local_project_directory = {}  # type: Dict[str, Pin]
        requirement_by_local_project_directory = {}  # type: Dict[str, Requirement]
        for locked_resolve in locked_resolves:
            for locked_requirement in locked_resolve.locked_requirements:
                if isinstance(locked_requirement.artifact, LocalProjectArtifact):
                    local_directory = locked_requirement.artifact.directory
                    local_pin = locked_requirement.pin
                    pin_by_local_project_directory[local_directory] = local_pin
                    requirement_by_local_project_directory[
                        local_directory
                    ] = local_pin.as_requirement()

        def extract_requirement(req):
            # type: (Union[Requirement, ParsedRequirement]) -> Requirement
            if isinstance(req, Requirement):
                return req
            if isinstance(req, LocalProjectRequirement):
                local_project_directory = os.path.abspath(req.path)
                pin = pin_by_local_project_directory[local_project_directory]
                requirement = Requirement.parse(
                    "{project_name}{extras}=={version}{marker}".format(
                        project_name=pin.project_name,
                        extras="[{extras}]".format(extras=",".join(req.extras))
                        if req.extras
                        else "",
                        version=pin.version,
                        marker="; {marker}".format(marker=req.marker) if req.marker else "",
                    )
                )
                # N.B.: We've already mapped all available local projects above, but the user may
                # have supplied the local project requirement with more specific constraints (
                # extras and / or marker restrictions) and we need to honor those; so we over-write.
                requirement_by_local_project_directory[local_project_directory] = requirement
                return requirement
            return req.requirement

        resolve_requirements = OrderedSet(extract_requirement(req) for req in requirements)

        return cls(
            pex_version=pex_version,
            style=style,
            requires_python=SortedTuple(requires_python),
            target_systems=SortedTuple(target_systems),
            pip_version=pip_version,
            resolver_version=resolver_version,
            requirements=SortedTuple(resolve_requirements, key=str),
            constraints=SortedTuple(constraints, key=str),
            allow_prereleases=allow_prereleases,
            allow_wheels=allow_wheels,
            allow_builds=allow_builds,
            prefer_older_binary=prefer_older_binary,
            use_pep517=use_pep517,
            build_isolation=build_isolation,
            transitive=transitive,
            locked_resolves=SortedTuple(locked_resolves),
            local_project_requirement_mapping=requirement_by_local_project_directory,
            source=source,
        )

    pex_version = attr.ib()  # type: str
    style = attr.ib()  # type: LockStyle.Value
    requires_python = attr.ib()  # type: SortedTuple[str]
    target_systems = attr.ib()  # type: SortedTuple[TargetSystem.Value]
    pip_version = attr.ib()  # type: PipVersionValue
    resolver_version = attr.ib()  # type: ResolverVersion.Value
    requirements = attr.ib()  # type: SortedTuple[Requirement]
    constraints = attr.ib()  # type: SortedTuple[Requirement]
    allow_prereleases = attr.ib()  # type: bool
    allow_wheels = attr.ib()  # type: bool
    allow_builds = attr.ib()  # type: bool
    prefer_older_binary = attr.ib()  # type: bool
    use_pep517 = attr.ib()  # type: Optional[bool]
    build_isolation = attr.ib()  # type: bool
    transitive = attr.ib()  # type: bool
    locked_resolves = attr.ib()  # type: SortedTuple[LockedResolve]
    local_project_requirement_mapping = attr.ib(eq=False)  # type: Mapping[str, Requirement]
    source = attr.ib(default=None, eq=False)  # type: Optional[str]



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os
import zipfile

from pex import layout, pex_warnings, variables
from pex.common import can_write_dir, open_zip, safe_mkdtemp
from pex.compatibility import PY2, WINDOWS
from pex.compatibility import string as compatibility_string
from pex.inherit_path import InheritPath
from pex.orderedset import OrderedSet
from pex.typing import TYPE_CHECKING, cast
from pex.variables import ENV, Variables
from pex.venv.bin_path import BinPath
from pex.version import __version__ as pex_version

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Mapping, Optional, Text, Tuple, Union

    # N.B.: These are expensive imports and PexInfo is used during PEX bootstrapping which we want
    # to be as fast as possible.
    from pex.interpreter import PythonInterpreter
    from pex.interpreter_constraints import InterpreterConstraints


# TODO(wickman) Split this into a PexInfoBuilder/PexInfo to ensure immutability.
# Issue #92.
class PexInfo(object):
    """PEX metadata.

    # Build metadata:
    build_properties: BuildProperties  # (key-value information about the build system)
    code_hash: str                     # sha1 hash of all names/code in the archive
    distributions: {dist_name: str}    # map from distribution name (i.e. path in
                                       # the internal cache) to its cache key (sha1)
    pex_hash: str                      # sha1 hash of all names/code and distributions in the pex
    requirements: list                 # list of requirements for this environment

    # Environment options
    pex_root: string                    # root of all pex-related files eg: ~/.pex
    entry_point: string                 # entry point into this pex
    script: string                      # script to execute in this pex environment
                                        # at most one of script/entry_point can be specified
    inherit_path: false/fallback/prefer # should this pex inherit site-packages + user site-packages
                                        # + PYTHONPATH?
    ignore_errors: True, default False  # should we ignore inability to resolve dependencies?

    .. versionchanged:: 0.8
      Removed the ``repositories`` and ``indices`` information, as they were never
      implemented.
    """

    PATH = layout.PEX_INFO_PATH
    BOOTSTRAP_CACHE = "bootstraps"
    INSTALL_CACHE = "installed_wheels"

    @classmethod
    def make_build_properties(cls):
        return {
            "pex_version": pex_version,
        }

    @classmethod
    def default(cls):
        # type: () -> PexInfo
        return cls(info={"build_properties": cls.make_build_properties()})

    @classmethod
    def from_pex(cls, pex):
        # type: (str) -> PexInfo
        if zipfile.is_zipfile(pex):  # Zip App PEX
            with open_zip(pex) as zf:
                pex_info = zf.read(cls.PATH)
        elif os.path.isfile(pex):  # Venv PEX
            with open(os.path.join(os.path.dirname(pex), cls.PATH), "rb") as fp:
                pex_info = fp.read()
        else:  # Directory (Either loose or installed) PEX
            with open(os.path.join(pex, cls.PATH), "rb") as fp:
                pex_info = fp.read()
        return cls.from_json(pex_info)

    @classmethod
    def from_json(cls, content):
        # type: (Union[bytes, Text]) -> PexInfo
        if isinstance(content, bytes):
            content = content.decode("utf-8")
        return cls(info=json.loads(content))

    @classmethod
    def from_env(cls, env=ENV):
        # type: (Variables) -> PexInfo
        pex_inherit_path = Variables.PEX_INHERIT_PATH.strip_default(env)
        inherit_path = None if pex_inherit_path is None else pex_inherit_path.value

        pex_info = {
            "pex_root": Variables.PEX_ROOT.strip_default(env),
            "entry_point": env.PEX_MODULE,
            "script": env.PEX_SCRIPT,
            "venv": Variables.PEX_VENV.strip_default(env),
            "inherit_path": inherit_path,
            "ignore_errors": Variables.PEX_IGNORE_ERRORS.strip_default(env),
        }
        # Filter out empty entries not explicitly set in the environment.
        return cls(info={k: v for k, v in pex_info.items() if v is not None})

    @classmethod
    def _parse_requirement_tuple(cls, requirement_tuple):
        if isinstance(requirement_tuple, (tuple, list)):
            if len(requirement_tuple) != 3:
                raise ValueError("Malformed PEX requirement: %r" % (requirement_tuple,))
            # pre 0.8.x requirement type:
            pex_warnings.warn(
                "Attempting to use deprecated PEX feature.  Please upgrade past PEX 0.8.x."
            )
            return requirement_tuple[0]
        elif isinstance(requirement_tuple, compatibility_string):
            return requirement_tuple
        raise ValueError("Malformed PEX requirement: %r" % (requirement_tuple,))

    def __init__(self, info=None):
        # type: (Optional[Mapping[str, Any]]) -> None
        """Construct a new PexInfo.

        This should not be used directly.
        """

        if info is not None and not isinstance(info, dict):
            raise ValueError(
                "PexInfo can only be seeded with a dict, got: " "%s of type %s" % (info, type(info))
            )
        self._pex_info = dict(info) if info else {}  # type: Dict[str, Any]
        self._distributions = self._pex_info.get("distributions", {})  # type: Dict[str, str]

        # N.B.: InterpreterConstraints is an expensive type to import; so we do so lazily since
        # PexInfo is used during PEX bootstrapping which we want to be fast.
        self._interpreter_constraints = None  # type: Optional[InterpreterConstraints]

        requirements = self._pex_info.get("requirements", [])
        if not isinstance(requirements, (list, tuple)):
            raise ValueError("Expected requirements to be a list, got %s" % type(requirements))
        self._requirements = OrderedSet(self._parse_requirement_tuple(req) for req in requirements)

    def _get_safe(self, key):
        if key not in self._pex_info:
            return None
        value = self._pex_info[key]
        return value.encode("utf-8") if PY2 else value

    @property
    def build_properties(self):
        """Information about the system on which this PEX was generated.

        :returns: A dictionary containing metadata about the environment used to build this PEX.
        """
        return self._pex_info.get("build_properties", {})

    @build_properties.setter
    def build_properties(self, value):
        if not isinstance(value, dict):
            raise TypeError("build_properties must be a dictionary!")
        self._pex_info["build_properties"] = self.make_build_properties()
        self._pex_info["build_properties"].update(value)

    @property
    def inject_env(self):
        # type: () -> Dict[str, str]
        return dict(self._pex_info.get("inject_env", {}))

    @inject_env.setter
    def inject_env(self, value):
        # type: (Mapping[str, str]) -> None
        self._pex_info["inject_env"] = dict(value)

    @property
    def inject_args(self):
        # type: () -> Tuple[str, ...]
        return tuple(self._pex_info.get("inject_args", ()))

    @inject_args.setter
    def inject_args(self, value):
        # type: (Iterable[str]) -> None
        self._pex_info["inject_args"] = tuple(value)

    @property
    def venv(self):
        # type: () -> bool
        """Whether or not PEX should be converted to a venv before it's executed.

        Creating a venv from a PEX is a operation that can be cached on the 1st run of a given PEX
        file which results in lower startup latency in subsequent runs.
        """
        return self._pex_info.get("venv", False)

    @venv.setter
    def venv(self, value):
        # type: (bool) -> None
        self._pex_info["venv"] = bool(value)

    @property
    def venv_bin_path(self):
        # type: () -> BinPath.Value
        """When run as a venv, whether or not to include `bin/` scripts on the PATH."""
        return BinPath.for_value(self._pex_info.get("venv_bin_path", BinPath.FALSE.value))

    @venv_bin_path.setter
    def venv_bin_path(self, value):
        # type: (BinPath.Value) -> None
        self._pex_info["venv_bin_path"] = str(value)

    @property
    def venv_copies(self):
        # type: () -> bool
        return self._pex_info.get("venv_copies", False)

    @venv_copies.setter
    def venv_copies(self, value):
        # type: (bool) -> None
        self._pex_info["venv_copies"] = value

    @property
    def venv_site_packages_copies(self):
        # type: () -> bool
        return self._pex_info.get("venv_site_packages_copies", False)

    @venv_site_packages_copies.setter
    def venv_site_packages_copies(self, value):
        # type: (bool) -> None
        self._pex_info["venv_site_packages_copies"] = value

    @property
    def venv_hermetic_scripts(self):
        # type: () -> bool
        return self._pex_info.get("venv_hermetic_scripts", True)

    @venv_hermetic_scripts.setter
    def venv_hermetic_scripts(self, value):
        # type: (bool) -> None
        self._pex_info["venv_hermetic_scripts"] = value

    def _venv_dir(
        self,
        pex_root,  # type: str
        pex_file,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
        expand_pex_root=True,  # type: bool
    ):
        # type: (...) -> Optional[str]
        if not self.venv:
            return None
        if self.pex_hash is None:
            raise ValueError("The venv_dir was requested but no pex_hash was set.")
        return variables.venv_dir(
            pex_file=pex_file,
            pex_root=pex_root,
            pex_hash=self.pex_hash,
            has_interpreter_constraints=self.has_interpreter_constraints,
            interpreter=interpreter,
            pex_path=self.pex_path,
            expand_pex_root=expand_pex_root,
        )

    def runtime_venv_dir(
        self,
        pex_file,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> Optional[str]
        return self._venv_dir(self.pex_root, pex_file, interpreter)

    def raw_venv_dir(
        self,
        pex_file,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> Optional[str]
        """Distiguished from ``venv_dir`` by use of the raw_pex_root.
        We don't expand the pex_root at build time in case the pex_root is not
        writable or doesn't exist at build time.
        """
        return self._venv_dir(self.raw_pex_root, pex_file, interpreter, expand_pex_root=False)

    @property
    def includes_tools(self):
        # type: () -> bool
        return self._pex_info.get("includes_tools", self.venv)

    @includes_tools.setter
    def includes_tools(self, value):
        # type: (bool) -> None
        self._pex_info["includes_tools"] = bool(value)

    @property
    def strip_pex_env(self):
        """Whether or not this PEX should strip `PEX_*` env vars before executing its entrypoint.

        You might want to set this to `False` if this PEX executes other PEXes or the Pex CLI itself
        and you want the executed PEX to be controlled via PEX environment variables.
        """
        return self._pex_info.get("strip_pex_env", True)

    @strip_pex_env.setter
    def strip_pex_env(self, value):
        self._pex_info["strip_pex_env"] = bool(value)

    @property
    def pex_path(self):
        # type: () -> Tuple[str, ...]
        """A list of other pex files to merge into the runtime environment.

        This pex info property is used to persist the PEX_PATH environment variable into the pex
        info metadata for reuse within a built pex.
        """
        pex_paths = self._pex_info.get("pex_paths")
        if pex_paths:
            return tuple(cast("Iterable[str]", pex_paths))

        # Legacy PEX-INFO stored this in a single string as a colon-separated list.
        pex_path = self._pex_info.get("pex_path")
        if pex_path:
            return tuple(pex_path.split(":"))

        return ()

    @pex_path.setter
    def pex_path(self, value):
        # type: (Iterable[str]) -> None
        if not WINDOWS:
            # Store in the legacy format on the legacy supported OSes for backwards compatibility
            # with old tools reading new PEX-INFO.
            self._pex_info["pex_path"] = ":".join(value)
        self._pex_info["pex_paths"] = tuple(value)

    @property
    def inherit_path(self):
        # type: () -> InheritPath.Value
        """Whether or not this PEX should be allowed to inherit system dependencies.

        By default, PEX environments are scrubbed of all system distributions prior to execution.
        This means that PEX files cannot rely upon preexisting system libraries.

        By default inherit_path is false. This may be overridden at runtime by the $PEX_INHERIT_PATH
        environment variable.
        """
        inherit_path = self._pex_info.get("inherit_path")
        return InheritPath.for_value(inherit_path) if inherit_path else InheritPath.FALSE

    @inherit_path.setter
    def inherit_path(self, value):
        # type: (InheritPath.Value) -> None
        self._pex_info["inherit_path"] = value.value

    @property
    def has_interpreter_constraints(self):
        # type: () -> bool
        if self._interpreter_constraints is not None:
            return bool(self._interpreter_constraints)
        return bool(self._pex_info.get("interpreter_constraints", ()))

    @property
    def interpreter_constraints(self):
        # type: () -> InterpreterConstraints
        """A list of constraints that determine the interpreter compatibility for this pex, using
        the Requirement-style format, e.g. ``'CPython>=3', or just '>=2.7,<3'`` for requirements
        agnostic to interpreter class.

        This property will be used at exec time when bootstrapping a pex to search PEX_PYTHON_PATH
        for a list of compatible interpreters.
        """
        if self._interpreter_constraints is None:
            # N.B.: InterpreterConstraints is an expensive import and PexInfo is used during PEX
            # bootstrapping which we want to be as fast as possible; so we go through hoops to be
            # lazy here.
            from pex.interpreter_constraints import InterpreterConstraints

            self._interpreter_constraints = InterpreterConstraints.parse(
                *self._pex_info.get("interpreter_constraints", ())
            )
        return self._interpreter_constraints

    @interpreter_constraints.setter
    def interpreter_constraints(self, value):
        # type: (InterpreterConstraints) -> None
        self._interpreter_constraints = value

    @property
    def ignore_errors(self):
        return self._pex_info.get("ignore_errors", False)

    @ignore_errors.setter
    def ignore_errors(self, value):
        self._pex_info["ignore_errors"] = bool(value)

    @property
    def emit_warnings(self):
        return self._pex_info.get("emit_warnings", True)

    @emit_warnings.setter
    def emit_warnings(self, value):
        self._pex_info["emit_warnings"] = bool(value)

    @property
    def code_hash(self):
        # type: () -> Optional[str]
        return self._pex_info.get("code_hash")

    @code_hash.setter
    def code_hash(self, value):
        # type: (str) -> None
        self._pex_info["code_hash"] = value

    @property
    def pex_hash(self):
        # type: () -> Optional[str]
        return self._pex_info.get("pex_hash")

    @pex_hash.setter
    def pex_hash(self, value):
        # type: (str) -> None
        self._pex_info["pex_hash"] = value

    @property
    def entry_point(self):
        return self._get_safe("entry_point")

    @entry_point.setter
    def entry_point(self, value):
        self._pex_info["entry_point"] = value

    @property
    def script(self):
        return self._get_safe("script")

    @script.setter
    def script(self, value):
        self._pex_info["script"] = value

    def add_requirement(self, requirement):
        self._requirements.add(str(requirement))

    @property
    def requirements(self):
        return self._requirements

    def add_distribution(self, location, sha):
        self._distributions[location] = sha

    @property
    def distributions(self):
        # type: () -> Dict[str, str]
        return self._distributions

    @property
    def raw_pex_root(self):
        # type: () -> str
        return cast(str, self._pex_info.get("pex_root", os.path.join("~", ".pex")))

    @property
    def pex_root(self):
        # type: () -> str
        pex_root = os.path.expanduser(self.raw_pex_root)
        if not can_write_dir(pex_root):
            tmp_root = safe_mkdtemp()
            pex_warnings.warn(
                "PEX_ROOT is configured as {pex_root} but that path is un-writeable, "
                "falling back to a temporary PEX_ROOT of {tmp_root} which will hurt "
                "performance.".format(pex_root=pex_root, tmp_root=tmp_root)
            )
            pex_root = self._pex_info["pex_root"] = tmp_root
        return pex_root

    @pex_root.setter
    def pex_root(self, value):
        # type: (Optional[str]) -> None
        if value is None:
            self._pex_info.pop("pex_root", None)
        else:
            self._pex_info["pex_root"] = value

    @property
    def bootstrap_hash(self):
        # type: () -> Optional[str]
        return self._pex_info.get("bootstrap_hash")

    @bootstrap_hash.setter
    def bootstrap_hash(self, value):
        # type: (str) -> None
        self._pex_info["bootstrap_hash"] = value

    @property
    def bootstrap(self):
        # type: () -> str
        return layout.BOOTSTRAP_DIR

    @property
    def bootstrap_cache(self):
        # type: () -> Optional[str]
        if self.bootstrap_hash is None:
            return None
        return os.path.join(self.pex_root, self.BOOTSTRAP_CACHE, self.bootstrap_hash)

    @property
    def internal_cache(self):
        # type: () -> str
        return layout.DEPS_DIR

    @property
    def install_cache(self):
        return os.path.join(self.pex_root, self.INSTALL_CACHE)

    @property
    def zip_unsafe_cache(self):
        #: type: () -> str
        return os.path.join(self.pex_root, "user_code")

    def update(self, other):
        # type: (PexInfo) -> None
        if not isinstance(other, PexInfo):
            raise TypeError("Cannot merge a %r with PexInfo" % type(other))
        self._pex_info.update(other._pex_info)
        self._distributions.update(other.distributions)
        if other.has_interpreter_constraints:
            self._interpreter_constraints = self.interpreter_constraints.merged(
                other.interpreter_constraints
            )
        self._requirements.update(other.requirements)

    def as_json_dict(self):
        # type: () -> Dict[str, Any]
        data = self._pex_info.copy()
        data["inherit_path"] = self.inherit_path.value
        data["requirements"] = list(self._requirements)
        data["interpreter_constraints"] = [str(ic) for ic in self.interpreter_constraints]
        data["distributions"] = self._distributions.copy()
        return data

    def dump(self):
        # type: (...) -> str
        data = self.as_json_dict()
        data["requirements"].sort()
        data["interpreter_constraints"].sort()
        return json.dumps(data, sort_keys=True)

    def copy(self):
        # type: () -> PexInfo
        return PexInfo(self.as_json_dict())

    def merge_pex_path(self, pex_path):
        # type: (Iterable[str]) -> None
        """Merges a new PEX_PATH definition into the existing one (if any).

        :param pex_path: The PEX_PATH to merge.
        """
        if not pex_path:
            return
        merged_pex_path = OrderedSet(self.pex_path)
        merged_pex_path.update(pex_path)
        self.pex_path = tuple(merged_pex_path)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self._pex_info)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_confounding_site_packages_directory(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    local_app_data = os.path.join(str(tmpdir), "local_app_data")
    result = run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "python-certifi-win32==1.6.1",
            "--",
            "-c",
            "import certifi_win32; print(certifi_win32.__file__)",
        ],
        # N.B.: certifi_win32 requires `LOCALAPPDATA` be set in the env to import at all.
        env=make_env(LOCALAPPDATA=local_app_data),
    )
    result.assert_success()
    assert result.output.startswith(os.path.join(pex_root, "installed_wheels"))



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import tarfile
import warnings
from contextlib import contextmanager
from textwrap import dedent

import pytest

from pex.common import open_zip, temporary_dir
from pex.dist_metadata import (
    Distribution,
    MetadataError,
    ProjectNameAndVersion,
    Requirement,
    find_dist_info_file,
    project_name_and_version,
    requires_dists,
    requires_python,
)
from pex.pex_warnings import PEXWarning
from pex.pip.installation import get_pip
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Any, Iterator, Tuple


@contextmanager
def installed_wheel(wheel_path):
    # type: (str) -> Iterator[Distribution]
    with temporary_dir() as install_dir:
        get_pip().spawn_install_wheel(wheel=wheel_path, install_dir=install_dir).wait()
        yield Distribution.load(install_dir)


@contextmanager
def example_distribution(name):
    # type: (str) -> Iterator[Tuple[str, Distribution]]
    wheel_path = os.path.join("./tests/example_packages", name)
    with installed_wheel(wheel_path) as distribution:
        yield wheel_path, distribution


@contextmanager
def downloaded_sdist(requirement):
    # type: (str) -> Iterator[str]
    with temporary_dir() as td:
        download_dir = os.path.join(td, "download")
        get_pip().spawn_download_distributions(
            download_dir=download_dir,
            requirements=[requirement],
            transitive=False,
            use_wheel=False,
        ).wait()
        dists = os.listdir(download_dir)
        assert len(dists) == 1, "Expected 1 dist to be downloaded for {}.".format(requirement)
        sdist = os.path.join(download_dir, dists[0])
        assert sdist.endswith((".sdist", ".tar.gz", ".zip"))
        yield sdist


def as_requirement(project_name_and_version):
    # type: (ProjectNameAndVersion) -> str
    return "{}=={}".format(project_name_and_version.project_name, project_name_and_version.version)


PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION = ProjectNameAndVersion("pygoogleearth", "0.0.2")


@pytest.fixture(scope="module")
def pygoogleearth_zip_sdist():
    # type: () -> Iterator[str]
    with downloaded_sdist(as_requirement(PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION)) as sdist:
        assert sdist.endswith(".zip")
        yield sdist


PIP_PROJECT_NAME_AND_VERSION = ProjectNameAndVersion("pip", "20.3.1")


@pytest.fixture(scope="module")
def pip_tgz_sdist():
    # type: () -> Iterator[str]
    with downloaded_sdist(as_requirement(PIP_PROJECT_NAME_AND_VERSION)) as sdist:
        assert sdist.endswith(".tar.gz")
        yield sdist


@pytest.fixture(scope="module")
def pip_wheel(pip_tgz_sdist):
    # type: (str) -> Iterator[str]
    with temporary_dir() as wheel_dir:
        get_pip().spawn_build_wheels([pip_tgz_sdist], wheel_dir=wheel_dir).wait()
        wheels = os.listdir(wheel_dir)
        assert len(wheels) == 1, "Expected 1 wheel to be built for {}.".format(pip_tgz_sdist)
        wheel = os.path.join(wheel_dir, wheels[0])
        assert wheel.endswith(".whl")
        yield wheel


@pytest.fixture(scope="module")
def pip_distribution(pip_wheel):
    # type: (str) -> Iterator[Distribution]
    with installed_wheel(pip_wheel) as distribution:
        yield distribution


def test_project_name_and_version_from_filename(
    pygoogleearth_zip_sdist,  # type: str
    pip_tgz_sdist,  # type: str
    pip_wheel,  # type: str
):
    # type: (...) -> None
    assert PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION == ProjectNameAndVersion.from_filename(
        pygoogleearth_zip_sdist
    )
    assert PIP_PROJECT_NAME_AND_VERSION == ProjectNameAndVersion.from_filename(pip_tgz_sdist)
    assert PIP_PROJECT_NAME_AND_VERSION == ProjectNameAndVersion.from_filename(pip_wheel)


def test_project_name_and_version_from_filename_pep625():
    # type: () -> None
    assert ProjectNameAndVersion(
        "a-distribution-name", "1.2.3"
    ) == ProjectNameAndVersion.from_filename("a-distribution-name-1.2.3.sdist")


def test_project_name_and_version_from_filename_invalid():
    # type: () -> None
    with pytest.raises(MetadataError):
        ProjectNameAndVersion.from_filename("unknown_distribution.format")


def test_project_name_and_version_from_metadata(
    pygoogleearth_zip_sdist,  # type: str
    pip_tgz_sdist,  # type: str
    pip_wheel,  # type: str
    pip_distribution,  # type: Distribution
):
    # type: (...) -> None
    assert PYGOOGLEEARTH_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pygoogleearth_zip_sdist, fallback_to_filename=False
    )
    assert PIP_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pip_tgz_sdist, fallback_to_filename=False
    )
    assert PIP_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pip_wheel, fallback_to_filename=False
    )
    assert PIP_PROJECT_NAME_AND_VERSION == project_name_and_version(
        pip_distribution, fallback_to_filename=False
    )


def test_project_name_and_version_fallback(tmpdir):
    # type: (Any) -> None
    def tmp_path(relpath):
        # type: (str) -> str
        return os.path.join(str(tmpdir), relpath)

    expected_metadata_project_name_and_version = ProjectNameAndVersion("foo", "1.2.3")

    pkg_info_src = tmp_path("PKG-INFO")
    with open(pkg_info_src, "w") as fp:
        fp.write("Name: {}\n".format(expected_metadata_project_name_and_version.project_name))
        fp.write("Version: {}\n".format(expected_metadata_project_name_and_version.version))

    sdist_path = tmp_path("bar-baz-4.5.6.tar.gz")
    with tarfile.open(sdist_path, mode="w:gz") as tf:
        # N.B.: Valid PKG-INFO at an invalid location.
        tf.add(pkg_info_src, arcname="PKG-INFO")

    with ENV.patch(PEX_EMIT_WARNINGS="True"), warnings.catch_warnings(record=True) as events:
        assert project_name_and_version(sdist_path, fallback_to_filename=False) is None
        assert 1 == len(events)
        warning = events[0]
        assert PEXWarning == warning.category
        assert "bar-baz-4.5.6/PKG-INFO" in str(warning.message)

    assert ProjectNameAndVersion("bar-baz", "4.5.6") == project_name_and_version(
        sdist_path, fallback_to_filename=True
    )

    name_and_version = "eggs-7.8.9"
    pkf_info_path = "{}/PKG-INFO".format(name_and_version)

    def write_sdist_tgz(extension):
        sdist_path = tmp_path("{}.{}".format(name_and_version, extension))
        with tarfile.open(sdist_path, mode="w:gz") as tf:
            tf.add(pkg_info_src, arcname=pkf_info_path)
        return sdist_path

    assert expected_metadata_project_name_and_version == project_name_and_version(
        write_sdist_tgz("tar.gz"), fallback_to_filename=False
    )
    assert expected_metadata_project_name_and_version == project_name_and_version(
        write_sdist_tgz("sdist"), fallback_to_filename=False
    )

    zip_sdist_path = tmp_path("{}.zip".format(name_and_version))
    with open_zip(zip_sdist_path, mode="w") as zf:
        zf.write(pkg_info_src, arcname=pkf_info_path)

    assert expected_metadata_project_name_and_version == project_name_and_version(
        zip_sdist_path, fallback_to_filename=False
    )


def test_requires_python(
    pip_tgz_sdist,  # type: str
    pip_wheel,  # type: str
    pip_distribution,  # type: Distribution
):
    # type: (...) -> None
    expected_requires_python = SpecifierSet(">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*")
    assert expected_requires_python == requires_python(pip_tgz_sdist)
    assert expected_requires_python == requires_python(pip_wheel)
    assert expected_requires_python == requires_python(pip_distribution)


def test_requires_python_none(pygoogleearth_zip_sdist):
    # type: (str) -> None
    assert requires_python(pygoogleearth_zip_sdist) is None
    with example_distribution("aws_cfn_bootstrap-1.4-py2-none-any.whl") as (wheel_path, dist):
        assert requires_python(wheel_path) is None
        assert requires_python(dist) is None


def test_requires_dists():
    # type: () -> None
    with example_distribution("aws_cfn_bootstrap-1.4-py2-none-any.whl") as (wheel_path, dist):
        expected_requirements = [
            Requirement.parse(req)
            for req in ("python-daemon>=1.5.2,<2.0", "pystache>=0.4.0", "setuptools")
        ]
        assert expected_requirements == list(requires_dists(wheel_path))
        assert expected_requirements == list(requires_dists(dist))


def test_requires_dists_none(pygoogleearth_zip_sdist):
    # type: (str) -> None
    assert [] == list(requires_dists(pygoogleearth_zip_sdist))
    with example_distribution("MarkupSafe-1.0-cp27-cp27mu-linux_x86_64.whl") as (wheel_path, dist):
        assert [] == list(requires_dists(wheel_path))
        assert [] == list(requires_dists(dist))

    # This tests a strange case detailed here:
    #   https://github.com/pantsbuild/pex/issues/1201#issuecomment-791715585
    with downloaded_sdist("et-xmlfile==1.0.1") as sdist, warnings.catch_warnings(
        record=True
    ) as events:
        assert [] == list(requires_dists(sdist))
        assert len(events) == 1
        warning = events[0]
        assert PEXWarning == warning.category
        assert (
            dedent(
                """\
                Ignoring 1 `Requires` field in {sdist} metadata:
                1.) Requires: python (>=2.6.0)

                You may have issues using the 'et_xmlfile' distribution as a result.
                More information on this workaround can be found here:
                  https://github.com/pantsbuild/pex/issues/1201#issuecomment-791715585
                """
            ).format(sdist=sdist)
            == str(warning.message)
        )


def test_wheel_metadata_project_name_fuzzy_issues_1375():
    # type: () -> None
    with example_distribution("PyAthena-1.9.0-py2.py3-none-any.whl") as (wheel_path, dist):
        expected = ProjectNameAndVersion("PyAthena", "1.9.0")
        assert expected == project_name_and_version(wheel_path)
        assert expected == project_name_and_version(dist)

    with example_distribution("PyAthena-1.11.5-py2.py3-none-any.whl") as (wheel_path, dist):
        expected = ProjectNameAndVersion("pyathena", "1.11.5")
        assert expected == project_name_and_version(wheel_path)
        assert expected == project_name_and_version(dist)


def test_find_dist_info_file():
    # type: () -> None
    assert (
        find_dist_info_file(
            project_name="foo",
            version="1.0",
            filename="bar",
            listing=[],
        )
        is None
    )

    assert (
        find_dist_info_file(
            project_name="foo",
            version="1.0",
            filename="bar",
            listing=[
                "foo-1.0.dist-info/baz",
            ],
        )
        is None
    )

    assert "Foo-1.0.dist-info/bar" == find_dist_info_file(
        project_name="foo",
        version="1.0",
        filename="bar",
        listing=[
            "foo-100.dist-info/bar",
            "Foo-1.0.dist-info/bar",
            "foo-1.0.dist-info/bar",
        ],
    )

    assert "stress__-.-__Test-1.0rc0.dist-info/direct_url.json" == find_dist_info_file(
        project_name="Stress-.__Test",
        version="1.0rc0",
        filename="direct_url.json",
        listing=[
            "direct_url.json",
            "foo-1.0rc0.dist-info/direct_url.json",
            "stress__-.-__Test-1.0rc0.dist-info/direct_url.json",
        ],
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import functools
import os.path
import shutil
from collections import OrderedDict
from multiprocessing.pool import ThreadPool

from pex import resolver
from pex.atomic_directory import FileLockStyle
from pex.auth import PasswordDatabase, PasswordEntry
from pex.common import pluralize
from pex.compatibility import cpu_count
from pex.network_configuration import NetworkConfiguration
from pex.orderedset import OrderedSet
from pex.pep_503 import ProjectName
from pex.pip.local_project import digest_local_project
from pex.pip.tool import PackageIndexConfiguration
from pex.pip.vcs import digest_vcs_archive
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve.downloads import ArtifactDownloader
from pex.resolve.locked_resolve import (
    DownloadableArtifact,
    FileArtifact,
    LocalProjectArtifact,
    VCSArtifact,
)
from pex.resolve.lockfile.download_manager import DownloadedArtifact, DownloadManager
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.lockfile.subset import subset
from pex.resolve.requirement_configuration import RequirementConfiguration
from pex.resolve.resolver_configuration import ResolverVersion
from pex.resolve.resolvers import MAX_PARALLEL_DOWNLOADS, Installed, Resolver
from pex.resolver import BuildAndInstallRequest, BuildRequest, InstallRequest
from pex.result import Error, catch, try_
from pex.targets import Target, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, Iterable, Mapping, Optional, Sequence, Tuple, Union

    from pex.hashing import HintedDigest


class FileArtifactDownloadManager(DownloadManager[FileArtifact]):
    def __init__(
        self,
        file_lock_style,  # type: FileLockStyle.Value
        downloader,  # type: ArtifactDownloader
        pex_root=None,  # type: Optional[str]
    ):
        super(FileArtifactDownloadManager, self).__init__(
            pex_root=pex_root, file_lock_style=file_lock_style
        )
        self._downloader = downloader

    def save(
        self,
        artifact,  # type: FileArtifact
        project_name,  # type: ProjectName
        dest_dir,  # type: str
        digest,  # type: HintedDigest
    ):
        # type: (...) -> Union[str, Error]
        return self._downloader.download(artifact=artifact, dest_dir=dest_dir, digest=digest)


class VCSArtifactDownloadManager(DownloadManager[VCSArtifact]):
    def __init__(
        self,
        target,  # type: Target
        file_lock_style,  # type: FileLockStyle.Value
        indexes=None,  # type: Optional[Sequence[str]]
        find_links=None,  # type: Optional[Sequence[str]]
        resolver_version=None,  # type: Optional[ResolverVersion.Value]
        network_configuration=None,  # type: Optional[NetworkConfiguration]
        password_entries=(),  # type: Iterable[PasswordEntry]
        cache=None,  # type: Optional[str]
        use_pep517=None,  # type: Optional[bool]
        build_isolation=True,  # type: bool
        pex_root=None,  # type: Optional[str]
        pip_version=PipVersion.VENDORED,  # type: PipVersionValue
        resolver=None,  # type: Optional[Resolver]
    ):
        super(VCSArtifactDownloadManager, self).__init__(
            pex_root=pex_root, file_lock_style=file_lock_style
        )
        self._target = target
        self._indexes = indexes
        self._find_links = find_links
        self._resolver_version = resolver_version
        self._network_configuration = network_configuration
        self._password_entries = password_entries
        self._cache = cache
        self._use_pep517 = use_pep517
        self._build_isolation = build_isolation
        self._pip_version = pip_version
        self._resolver = resolver

    def save(
        self,
        artifact,  # type: VCSArtifact
        project_name,  # type: ProjectName
        dest_dir,  # type: str
        digest,  # type: HintedDigest
    ):
        # type: (...) -> Union[str, Error]

        requirement = artifact.as_unparsed_requirement(project_name)
        downloaded_vcs = resolver.download(
            targets=Targets.from_target(self._target),
            requirements=[requirement],
            transitive=False,
            indexes=self._indexes,
            find_links=self._find_links,
            resolver_version=self._resolver_version,
            network_configuration=self._network_configuration,
            password_entries=self._password_entries,
            use_wheel=False,
            prefer_older_binary=False,
            use_pep517=self._use_pep517,
            build_isolation=self._build_isolation,
            max_parallel_jobs=1,
            pip_version=self._pip_version,
            resolver=self._resolver,
        )
        if len(downloaded_vcs.local_distributions) != 1:
            return Error(
                "Expected 1 artifact for an intransitive download of {requirement}, found "
                "{count}:\n"
                "{downloads}".format(
                    requirement=requirement,
                    count=len(downloaded_vcs.local_distributions),
                    downloads="\n".join(
                        "{index}. {download}".format(index=index, download=download.path)
                        for index, download in enumerate(
                            downloaded_vcs.local_distributions, start=1
                        )
                    ),
                )
            )

        local_distribution = downloaded_vcs.local_distributions[0]
        filename = os.path.basename(local_distribution.path)
        digest_vcs_archive(
            archive_path=local_distribution.path,
            vcs=artifact.vcs,
            digest=digest,
        )
        shutil.move(local_distribution.path, os.path.join(dest_dir, filename))
        return filename


class LocalProjectDownloadManager(DownloadManager[LocalProjectArtifact]):
    def __init__(
        self,
        target,  # type: Target
        file_lock_style,  # type: FileLockStyle.Value
        pip_version,  # type: PipVersionValue
        resolver,  # type: Resolver
        pex_root=None,  # type: Optional[str]
    ):
        super(LocalProjectDownloadManager, self).__init__(
            pex_root=pex_root, file_lock_style=file_lock_style
        )
        self._target = target
        self._pip_version = pip_version
        self._resolver = resolver

    def save(
        self,
        artifact,  # type: LocalProjectArtifact
        project_name,  # type: ProjectName
        dest_dir,  # type: str
        digest,  # type: HintedDigest
    ):
        # type: (...) -> Union[str, Error]
        source_dir_or_error = digest_local_project(
            directory=artifact.directory,
            digest=digest,
            pip_version=self._pip_version,
            target=self._target,
            resolver=self._resolver,
            dest_dir=dest_dir,
        )
        if isinstance(source_dir_or_error, Error):
            return source_dir_or_error
        return os.path.basename(source_dir_or_error)


def download_artifact(
    downloadable_artifact_and_target,  # type: Tuple[DownloadableArtifact, Target]
    file_download_managers_by_target,  # type: Mapping[Target, FileArtifactDownloadManager]
    vcs_download_managers_by_target,  # type: Mapping[Target, VCSArtifactDownloadManager]
    local_project_download_managers_by_target,  # type: Mapping[Target, LocalProjectDownloadManager]
):
    # type: (...) -> Union[DownloadedArtifact, Error]
    downloadable_artifact, target = downloadable_artifact_and_target
    if isinstance(downloadable_artifact.artifact, VCSArtifact):
        return catch(
            vcs_download_managers_by_target[target].store,
            downloadable_artifact.artifact,
            downloadable_artifact.pin.project_name,
        )

    if isinstance(downloadable_artifact.artifact, FileArtifact):
        return catch(
            file_download_managers_by_target[target].store,
            downloadable_artifact.artifact,
            downloadable_artifact.pin.project_name,
        )

    return catch(
        local_project_download_managers_by_target[target].store,
        downloadable_artifact.artifact,
        downloadable_artifact.pin.project_name,
    )


def resolve_from_lock(
    targets,  # type: Targets
    lock,  # type: Lockfile
    resolver,  # type: Resolver
    requirements=None,  # type: Optional[Iterable[str]]
    requirement_files=None,  # type: Optional[Iterable[str]]
    constraint_files=None,  # type: Optional[Iterable[str]]
    indexes=None,  # type: Optional[Sequence[str]]
    find_links=None,  # type: Optional[Sequence[str]]
    resolver_version=None,  # type: Optional[ResolverVersion.Value]
    network_configuration=None,  # type: Optional[NetworkConfiguration]
    password_entries=(),  # type: Iterable[PasswordEntry]
    build=True,  # type: bool
    use_wheel=True,  # type: bool
    prefer_older_binary=False,  # type: bool
    use_pep517=None,  # type: Optional[bool]
    build_isolation=True,  # type: bool
    compile=False,  # type: bool
    transitive=True,  # type: bool
    verify_wheels=True,  # type: bool
    max_parallel_jobs=None,  # type: Optional[int]
    pip_version=PipVersion.VENDORED,  # type: PipVersionValue
):
    # type: (...) -> Union[Installed, Error]

    subset_result = try_(
        subset(
            targets=targets,
            lock=lock,
            requirement_configuration=RequirementConfiguration(
                requirements=requirements,
                requirement_files=requirement_files,
                constraint_files=constraint_files,
            ),
            network_configuration=network_configuration,
            build=build,
            use_wheel=use_wheel,
            prefer_older_binary=prefer_older_binary,
            transitive=transitive,
        )
    )
    downloadable_artifacts_and_targets = OrderedSet(
        (downloadable_artifact, resolved_subset.target)
        for resolved_subset in subset_result.subsets
        for downloadable_artifact in resolved_subset.resolved.downloadable_artifacts
    )

    # Since the download managers are stored to via a thread pool, we need to use BSD style locks.
    # These locks are not as portable as POSIX style locks but work with threading unlike POSIX
    # locks which are subject to threading-unaware deadlock detection per the standard. Linux, in
    # fact, implements deadlock detection for POSIX locks; so we can run afoul of false EDEADLCK
    # errors under the right interleaving of processes and threads and download artifact targets.
    file_lock_style = FileLockStyle.BSD

    file_download_managers_by_target = {
        resolved_subset.target: FileArtifactDownloadManager(
            file_lock_style=file_lock_style,
            downloader=ArtifactDownloader(
                resolver=resolver,
                target=resolved_subset.target,
                package_index_configuration=PackageIndexConfiguration.create(
                    pip_version=pip_version,
                    resolver_version=resolver_version,
                    indexes=indexes,
                    find_links=find_links,
                    network_configuration=network_configuration,
                    password_entries=PasswordDatabase.from_netrc().append(password_entries).entries,
                ),
                max_parallel_jobs=max_parallel_jobs,
            ),
        )
        for resolved_subset in subset_result.subsets
    }

    vcs_download_managers_by_target = {
        resolved_subset.target: VCSArtifactDownloadManager(
            target=resolved_subset.target,
            file_lock_style=file_lock_style,
            indexes=indexes,
            find_links=find_links,
            resolver_version=resolver_version,
            network_configuration=network_configuration,
            password_entries=password_entries,
            use_pep517=use_pep517,
            build_isolation=build_isolation,
            pip_version=pip_version,
            resolver=resolver,
        )
        for resolved_subset in subset_result.subsets
    }

    local_project_download_managers_by_target = {
        resolved_subset.target: LocalProjectDownloadManager(
            file_lock_style=file_lock_style,
            pip_version=pip_version,
            target=resolved_subset.target,
            resolver=resolver,
        )
        for resolved_subset in subset_result.subsets
    }

    max_threads = min(
        len(downloadable_artifacts_and_targets) or 1,
        min(MAX_PARALLEL_DOWNLOADS, 4 * (max_parallel_jobs or cpu_count() or 1)),
    )
    with TRACER.timed(
        "Downloading {url_count} distributions to satisfy {requirement_count} requirements".format(
            url_count=len(downloadable_artifacts_and_targets),
            requirement_count=len(subset_result.requirements),
        )
    ):
        pool = ThreadPool(processes=max_threads)
        try:
            download_results = tuple(
                zip(
                    tuple(
                        downloadable_artifact
                        for downloadable_artifact, _ in downloadable_artifacts_and_targets
                    ),
                    pool.map(
                        functools.partial(
                            download_artifact,
                            file_download_managers_by_target=file_download_managers_by_target,
                            vcs_download_managers_by_target=vcs_download_managers_by_target,
                            local_project_download_managers_by_target=local_project_download_managers_by_target,
                        ),
                        downloadable_artifacts_and_targets,
                    ),
                )
            )
        finally:
            pool.close()
            pool.join()

    with TRACER.timed("Categorizing {} downloaded artifacts".format(len(download_results))):
        downloaded_artifacts = {}  # type: Dict[DownloadableArtifact, DownloadedArtifact]
        download_errors = OrderedDict()  # type: OrderedDict[DownloadableArtifact, Error]
        for downloadable_artifact, download_result in download_results:
            if isinstance(download_result, DownloadedArtifact):
                downloaded_artifacts[downloadable_artifact] = download_result
            else:
                download_errors[downloadable_artifact] = download_result

        if download_errors:
            error_count = len(download_errors)
            return Error(
                "There {were} {count} {errors} downloading required artifacts:\n"
                "{error_details}".format(
                    were="was" if error_count == 1 else "were",
                    count=error_count,
                    errors=pluralize(download_errors, "error"),
                    error_details="\n".join(
                        "{index}. {pin} from {url}\n    {error}".format(
                            index=index,
                            pin=downloadable_artifact.pin,
                            url=downloadable_artifact.artifact.url,
                            error=error,
                        )
                        for index, (downloadable_artifact, error) in enumerate(
                            download_errors.items(), start=1
                        )
                    ),
                )
            )

        build_requests = []
        install_requests = []
        for resolved_subset in subset_result.subsets:
            for downloadable_artifact in resolved_subset.resolved.downloadable_artifacts:
                downloaded_artifact = downloaded_artifacts[downloadable_artifact]
                if downloaded_artifact.path.endswith(".whl"):
                    install_requests.append(
                        InstallRequest(
                            target=resolved_subset.target,
                            wheel_path=downloaded_artifact.path,
                            fingerprint=downloaded_artifact.fingerprint,
                        )
                    )
                else:
                    build_requests.append(
                        BuildRequest(
                            target=resolved_subset.target,
                            source_path=downloaded_artifact.path,
                            fingerprint=downloaded_artifact.fingerprint,
                        )
                    )

    with TRACER.timed(
        "Building {} artifacts and installing {}".format(
            len(build_requests), len(build_requests) + len(install_requests)
        )
    ):
        build_and_install_request = BuildAndInstallRequest(
            build_requests=build_requests,
            install_requests=install_requests,
            direct_requirements=subset_result.requirements,
            package_index_configuration=PackageIndexConfiguration.create(
                pip_version=pip_version,
                resolver_version=resolver_version,
                indexes=indexes,
                find_links=find_links,
                network_configuration=network_configuration,
                password_entries=PasswordDatabase.from_netrc().append(password_entries).entries,
            ),
            compile=compile,
            prefer_older_binary=prefer_older_binary,
            use_pep517=use_pep517,
            build_isolation=build_isolation,
            verify_wheels=verify_wheels,
            pip_version=pip_version,
            resolver=resolver,
        )
        installed_distributions = build_and_install_request.install_distributions(
            # This otherwise checks that resolved distributions all meet internal requirement
            # constraints (This allows pip-legacy-resolver resolves with invalid solutions to be
            # failed post-facto by Pex at PEX build time). We've already done this via
            # `LockedResolve.resolve` above and need not waste time (~O(100ms)) doing this again.
            ignore_errors=True,
            max_parallel_jobs=max_parallel_jobs,
            local_project_directory_to_sdist={
                downloadable_artifact.artifact.directory: downloaded_artifact.path
                for downloadable_artifact, downloaded_artifact in downloaded_artifacts.items()
                if isinstance(downloadable_artifact.artifact, LocalProjectArtifact)
            },
        )
    return Installed(installed_distributions=tuple(installed_distributions))



# SPDX-License-Identifier: MIT

from __future__ import absolute_import, division, print_function

import sys

from functools import partial

from . import converters, exceptions, filters, setters, validators
from ._cmp import cmp_using
from ._config import get_run_validators, set_run_validators
from ._funcs import asdict, assoc, astuple, evolve, has, resolve_types
from ._make import (
    NOTHING,
    Attribute,
    Factory,
    attrib,
    attrs,
    fields,
    fields_dict,
    make_class,
    validate,
)
from ._version_info import VersionInfo


__version__ = "21.5.0.dev0"
__version_info__ = VersionInfo._from_version_string(__version__)

__title__ = "attrs"
__description__ = "Classes Without Boilerplate"
__url__ = "https://www.attrs.org/"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Hynek Schlawack"
__email__ = "hs@ox.cx"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2015 Hynek Schlawack"


s = attributes = attrs
ib = attr = attrib
dataclass = partial(attrs, auto_attribs=True)  # happy Easter ;)

__all__ = [
    "Attribute",
    "Factory",
    "NOTHING",
    "asdict",
    "assoc",
    "astuple",
    "attr",
    "attrib",
    "attributes",
    "attrs",
    "cmp_using",
    "converters",
    "evolve",
    "exceptions",
    "fields",
    "fields_dict",
    "filters",
    "get_run_validators",
    "has",
    "ib",
    "make_class",
    "resolve_types",
    "s",
    "set_run_validators",
    "setters",
    "validate",
    "validators",
]

if sys.version_info[:2] >= (3, 6):
    from ._next_gen import define, field, frozen, mutable  # noqa: F401

    __all__.extend(("define", "field", "frozen", "mutable"))



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.typing import TYPE_CHECKING, Generic, cast, overload

if TYPE_CHECKING:
    from typing import Any, Callable, ClassVar, Iterable, Iterator, Optional, Tuple, Type, TypeVar

    _I = TypeVar("_I")
    _R = TypeVar("_R", bound="Rank")


class Rank(Generic["_R"]):
    """Represents a ranking where lower values represent higher ranks.

    Ranks naturally sort from highest rank (lowest value) to lowest rank (highest value).

    Rank is intended to be sub-classed to define a ranking for a given set of subjects. Subclasses
    can adjust the value of the highest rank by over-riding the `HIGHEST_NATURAL_VALUE` class
    variable.
    """

    HIGHEST_NATURAL_VALUE = 1  # type: ClassVar[int]

    @classmethod
    def highest_natural(cls):
        # type: (Type[_R]) -> _R
        return cls(cls.HIGHEST_NATURAL_VALUE)

    @classmethod
    def ranked(
        cls,  # type: Type[_R]
        items,  # type: Iterable[_I]
    ):
        # type: (...) -> Iterator[Tuple[_I, _R]]
        for rank, item in enumerate(items, start=cls.HIGHEST_NATURAL_VALUE):
            yield item, cls(rank)

    @classmethod
    @overload
    def select_highest_rank(
        cls,
        item1,  # type: Rank
        item2,  # type: Rank
    ):
        # type: (...) -> Rank
        pass

    @classmethod
    @overload
    def select_highest_rank(
        cls,
        item1,  # type: _I
        item2,  # type: _I
        extract_rank,  # type: Callable[[_I], Rank]
    ):
        # type: (...) -> _I
        pass

    @classmethod
    def select_highest_rank(
        cls,
        item1,  # type: _I
        item2,  # type: _I
        extract_rank=None,  # type: Optional[Callable[[_I], Rank]]
    ):
        # type: (...) -> _I
        if extract_rank is not None:
            return item1 if extract_rank(item1) < extract_rank(item2) else item2

        if not isinstance(item1, Rank):
            raise TypeError(
                "Can only select from amongst the same rank type; item1 is a `{item1_type}` "
                "and not a `Rank`. You may need to supply an `extract_rank` function.".format(
                    item1_type=type(item1).__name__
                )
            )
        if type(item1) != type(item2):
            raise TypeError(
                "Can only select from amongst the same rank type; item1 is a `{item1_type}` "
                "but item2 is a `{item2_type}`.".format(
                    item1_type=item1.__class__.__name__, item2_type=type(item2).__name__
                )
            )
        return cast("_I", item1 if item1 < item2 else item2)

    def __init__(self, value):
        # type: (int) -> None
        self._value = value

    @property
    def value(self):
        # type: () -> int
        return self._value

    def higher(self):
        # type: (_R) -> _R
        return self.__class__(self._value - 1)

    def lower(self):
        # type: (_R) -> _R
        return self.__class__(self._value + 1)

    def __repr__(self):
        # type: () -> str
        return "{class_name}({value})".format(class_name=self.__class__.__name__, value=self._value)

    def __eq__(self, other):
        # type: (Any) -> bool
        return type(self) == type(other) and self._value == other._value

    def __ne__(self, other):
        # type: (Any) -> bool
        return not self == other

    def __lt__(self, other):
        # type: (Any) -> bool
        if type(self) != type(other):
            return NotImplemented
        # The need for this ignore appears to be tied to `return NotImplemented` but it seems that
        # MyPy doesn't understand `NotImplemented` as either a `Literal` value or as a type in
        # `Union`.
        return self._value < other.value  # type: ignore[no-any-return]



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import re
from io import BytesIO

try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore[no-redef,import]

import pytest

from pex.fetcher import URLFetcher
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.resolve.pep_691.api import Client
from pex.resolve.pep_691.model import Endpoint, File, Meta, Project
from pex.resolve.resolved_requirement import Fingerprint
from pex.sorted_tuple import SortedTuple
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict


def client_request(response_body):
    # type: (bytes) -> Project
    with mock.patch.object(URLFetcher, "get_body_stream", return_value=BytesIO(response_body)):
        return Client().request(
            Endpoint(
                url="https://example.org/simple/",
                content_type="application/vnd.pypi.simple.v1+json",
            )
        )


def test_request_nominal():
    # type: () -> None
    assert Project(
        name=ProjectName("p537"),
        files=SortedTuple(
            [
                File(
                    filename="file.tar.gz",
                    url="https://files.example.org/simple/file.tar.gz",
                    hashes=SortedTuple(
                        [
                            Fingerprint("md5", "weak"),
                            Fingerprint("sha256", "strong"),
                        ]
                    ),
                ),
            ]
        ),
        meta=Meta(api_version=Version("1.2")),
    ) == client_request(
        json.dumps(
            {
                "name": "p537",
                "files": [
                    {
                        "filename": "file.tar.gz",
                        "url": "https://files.example.org/simple/file.tar.gz",
                        "hashes": {"md5": "weak", "sha256": "strong"},
                    },
                ],
                "meta": {"api-version": "1.2"},
            }
        ).encode("utf-8")
    )


def test_request_url_absolutize():
    # type: () -> None
    assert Project(
        name=ProjectName("p537"),
        files=SortedTuple(
            [
                File(
                    filename="file.whl",
                    url="https://example.org/simple/__files/relative/file.whl",
                    hashes=SortedTuple([Fingerprint("md5", "collision")]),
                ),
            ]
        ),
        meta=Meta(api_version=Version("1.2")),
    ) == client_request(
        json.dumps(
            {
                "name": "p537",
                "files": [
                    {
                        "filename": "file.whl",
                        "url": "__files/relative/file.whl",
                        "hashes": {"md5": "collision"},
                    },
                ],
                "meta": {"api-version": "1.2"},
            }
        ).encode("utf-8")
    )


def test_bad_json_response():
    # type: () -> None
    with pytest.raises(
        Client.Error,
        match=r"^{}.*".format(
            re.escape(
                "PEP-691 API request to https://example.org/simple/ for "
                "application/vnd.pypi.simple.v1+json returned invalid JSON: "
            )
        ),
    ):
        client_request(b"[[]")


def serialize_response(data):
    # type: (Dict[str, Any]) -> bytes
    return json.dumps(data, indent=2).encode("utf-8")


def assert_response_error_starts_with(
    response,  # type: Dict[str, Any]
    prefix,  # type: str
):
    # type: (...) -> None

    with pytest.raises(Client.Error, match=r"^{}.*".format(re.escape(prefix))):
        client_request(serialize_response(response))


def test_unsupported_api_version():
    assert_response_error_starts_with(
        {"name": "p537", "files": [], "meta": {"api-version": "2.0"}},
        prefix=(
            "PEP-691 API response from https://example.org/simple/ for "
            "application/vnd.pypi.simple.v1+json reports an api-version of 2.0 and Pex "
            "currently only supports api-version 1.x:"
        ),
    )


@pytest.fixture
def valid_response():
    # type: () -> Dict[str, Any]
    return {
        "name": "p537",
        "files": [
            {
                "filename": "file.whl",
                "url": "__files/relative/file.whl",
                "hashes": {"md5": "collision"},
            }
        ],
        "meta": {"api-version": "1.2"},
    }


def test_missing_name(valid_response):
    # type: (Dict[str, Any]) -> None

    valid_response.pop("name")
    assert_response_error_starts_with(
        valid_response,
        prefix=(
            "PEP-691 API response from https://example.org/simple/ for "
            "application/vnd.pypi.simple.v1+json did not contain the expected key '.[\"name\"]':"
        ),
    )


def test_missing_meta_version(valid_response):
    # type: (Dict[str, Any]) -> None

    valid_response["meta"].pop("api-version")
    assert_response_error_starts_with(
        valid_response,
        prefix=(
            "PEP-691 API response from https://example.org/simple/ for "
            "application/vnd.pypi.simple.v1+json did not contain the expected key "
            "'.meta[\"api-version\"]':"
        ),
    )


def test_bad_hash(valid_response):
    # type: (Dict[str, Any]) -> None

    valid_response["files"][0]["hashes"]["sha256"] = 42
    assert_response_error_starts_with(
        valid_response,
        prefix=(
            "PEP-691 API response from https://example.org/simple/ for "
            "application/vnd.pypi.simple.v1+json reports a hash value of 42 of type int for "
            "'.files[0].hashes[\"sha256\"]' but hash values should be strings:"
        ),
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess

from pex.cli.testing import run_pex3
from pex.interpreter import PythonInterpreter
from pex.pex_info import PexInfo
from pex.resolve.lockfile import json_codec
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_multiplatform_sdist(
    tmpdir,  # type: Any
    py27,  # type: PythonInterpreter
    py38,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    all_interpreters = (py27, py38, py310)
    python_path = os.pathsep.join((interp.binary for interp in all_interpreters))
    interpreter_selection_args = [
        "--python-path",
        python_path,
        "--interpreter-constraint",
        ">=2.7,<3.11",
    ]

    lock = os.path.join(str(tmpdir), "lock")
    run_pex3(
        "lock",
        "create",
        "--style",
        "universal",
        "--no-wheel",
        "psutil==5.9.0",
        "-o",
        lock,
        *interpreter_selection_args
    ).assert_success()
    lock_file = json_codec.load(lock)
    assert 1 == len(lock_file.locked_resolves), "Expected 1 resolve for universal style."
    locked_resolve = lock_file.locked_resolves[0]
    assert 1 == len(
        locked_resolve.locked_requirements
    ), "Expected 1 locked requirement since psutil has no dependencies"
    locked_requirement = locked_resolve.locked_requirements[0]
    assert 0 == len(
        locked_requirement.additional_artifacts
    ), "Expected just a single sdist artifact since we specified --no-wheel."
    assert locked_requirement.artifact.url.endswith(".tar.gz"), "Expected a locked sdist URL."

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["--lock", lock, "-o", pex] + interpreter_selection_args).assert_success()

    assert 3 == len(
        PexInfo.from_pex(pex).distributions
    ), "Expected a unique platform-specific wheel to be built for each interpreter"
    for interp in all_interpreters:
        subprocess.check_call(args=[interp.binary, pex, "-c", "import psutil"])



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import itertools
import os.path

from pex.orderedset import OrderedSet
from pex.rank import Rank
from pex.third_party.packaging.tags import Tag, parse_tag
from pex.typing import TYPE_CHECKING, cast, overload

if TYPE_CHECKING:
    from typing import Iterable, Iterator, List, Mapping, MutableMapping, Optional, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


def _prepare_tags(tags):
    # type: (Iterable[Tag]) -> Tuple[Tag, ...]
    return tags if isinstance(tags, tuple) else tuple(OrderedSet(tags))


class TagRank(Rank["TagRank"]):
    """A Rank new-type to single out the ranking scheme for tags from other ranking schemes.

    The highest rank tag is the most specific tag.
    """


@attr.s(frozen=True)
class RankedTag(object):
    tag = attr.ib(order=False)  # type: Tag
    rank = attr.ib()  # type: TagRank

    def select_higher_rank(self, other):
        # type: (RankedTag) -> RankedTag
        return Rank.select_highest_rank(
            self, other, extract_rank=lambda ranked_tag: ranked_tag.rank
        )


@attr.s(frozen=True)
class CompatibilityTags(object):
    """A ranked set of PEP-425 compatibility tags.

    Tags are ordered most specific 1st to most generic last. The more specific a tag, the lower its
    rank value, with the most specific tag (best match) being ranked 0.

    See: https://www.python.org/dev/peps/pep-0425/#use
    """

    @classmethod
    def from_wheel(cls, wheel):
        # type: (str) -> CompatibilityTags
        wheel_stem, ext = os.path.splitext(os.path.basename(wheel))
        if ".whl" != ext:
            raise ValueError(
                "Can only calculate wheel tags from a filename that ends in .whl per "
                "https://peps.python.org/pep-0427/#file-name-convention, given: {wheel!r}".format(
                    wheel=wheel
                )
            )
        # Wheel filename format: https://www.python.org/dev/peps/pep-0427/#file-name-convention
        # `{distribution}-{version}(-{build tag})?-{python tag}-{abi tag}-{platform tag}.whl`
        wheel_components = wheel_stem.rsplit("-", 3)
        if len(wheel_components) != 4:
            pattern = "`-{python tag}-{abi tag}-{platform tag}.whl`"
            raise ValueError(
                "Can only calculate wheel tags from a filename that ends in {pattern} per "
                "https://peps.python.org/pep-0427/#file-name-convention, given: {wheel!r}".format(
                    pattern=pattern, wheel=wheel
                )
            )
        return cls(tags=tuple(parse_tag("-".join(wheel_components[-3:]))))

    @classmethod
    def from_strings(cls, tags):
        # type: (Iterable[str]) -> CompatibilityTags
        return cls(tags=tuple(itertools.chain.from_iterable(parse_tag(tag) for tag in tags)))

    _tags = attr.ib(converter=_prepare_tags)  # type: Tuple[Tag, ...]
    __rankings = attr.ib(eq=False, factory=dict)  # type: MutableMapping[Tag, TagRank]

    @_tags.validator
    def _validate_tags(
        self,
        attribute,  # type: attr.Attribute
        value,  # type: Tuple[Tag, ...]
    ):
        if not value:
            raise ValueError(
                "The {name} parameter should contain at least one tag; given an empty set.".format(
                    name=attribute.name
                )
            )

    def extend(self, tags):
        # type: (Iterable[Tag]) -> CompatibilityTags
        return CompatibilityTags(self._tags + tuple(tags))

    def compatible_tags(self, tags):
        # type: (Iterable[Tag]) -> OrderedSet[Tag]

        query = frozenset(tags)

        def iter_compatible():
            for tag in self:
                if tag in query:
                    yield tag

        return OrderedSet(iter_compatible())

    def to_string_list(self):
        # type: () -> List[str]
        return [str(tag) for tag in self._tags]

    @property
    def _rankings(self):
        # type: () -> Mapping[Tag, TagRank]
        if not self.__rankings:
            self.__rankings.update(TagRank.ranked(self._tags))
        return self.__rankings

    @property
    def lowest_rank(self):
        # type: () -> TagRank
        return cast(TagRank, self.rank(self[-1]))

    def rank(self, tag):
        # type: (Tag) -> Optional[TagRank]
        return self._rankings.get(tag)

    def best_match(self, tags):
        # type: (Iterable[Tag]) -> Optional[RankedTag]
        best_match = None  # type: Optional[RankedTag]
        for tag in tags:
            rank = self.rank(tag)
            if rank is None:
                continue
            ranked_tag = RankedTag(tag=tag, rank=rank)
            if best_match is None or ranked_tag is best_match.select_higher_rank(ranked_tag):
                best_match = ranked_tag
        return best_match

    def __iter__(self):
        # type: () -> Iterator[Tag]
        return iter(self._tags)

    def __len__(self):
        # type: () -> int
        return len(self._tags)

    @overload
    def __getitem__(self, index):
        # type: (int) -> Tag
        pass

    @overload
    def __getitem__(self, tag):
        # type: (Tag) -> TagRank
        pass

    def __getitem__(self, index_or_tag):
        # type: (Union[int, Tag]) -> Union[Tag, TagRank]
        """Retrieve tag by its rank or a tags rank.

        Ranks are 0-based with the 0-rank tag being the most specific (best match).
        """
        if isinstance(index_or_tag, Tag):
            return self._rankings[index_or_tag]
        else:
            return self._tags[index_or_tag]



import unicodedata
import sys

if "__PEX_UNVENDORED__" in __import__("os").environ:
  from setuptools.extern import six  # vendor:skip
else:
  from pex.third_party.setuptools.extern import six



# HFS Plus uses decomposed UTF-8
def decompose(path):
    if isinstance(path, six.text_type):
        return unicodedata.normalize('NFD', path)
    try:
        path = path.decode('utf-8')
        path = unicodedata.normalize('NFD', path)
        path = path.encode('utf-8')
    except UnicodeError:
        pass  # Not UTF-8
    return path


def filesys_decode(path):
    """
    Ensure that the given path is decoded,
    NONE when no expected encoding works
    """

    if isinstance(path, six.text_type):
        return path

    fs_enc = sys.getfilesystemencoding() or 'utf-8'
    candidates = fs_enc, 'utf-8'

    for enc in candidates:
        try:
            return path.decode(enc)
        except UnicodeDecodeError:
            continue


def try_encode(string, enc):
    "turn unicode encoding into a functional routine"
    try:
        return string.encode(enc)
    except UnicodeEncodeError:
        return None



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import contextlib
import marshal
import os
import sys

import pytest

from pex import compatibility
from pex.common import safe_open, temporary_dir
from pex.compatibility import to_bytes
from pex.compiler import Compiler
from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, Iterable, Iterator, List, Text, Tuple


def write_source(path, valid=True):
    # type: (str, bool) -> None
    with safe_open(path, "wb") as fp:
        fp.write(to_bytes("basename = %r\n" % os.path.basename(path)))
        if not valid:
            fp.write(to_bytes("invalid!\n"))


@contextlib.contextmanager
def compilation(valid_paths, invalid_paths, compile_paths):
    # type: (Iterable[str], Iterable[str], Iterable[str]) -> Iterator[Tuple[str, List[Text]]]
    with temporary_dir() as root:
        for path in valid_paths:
            write_source(os.path.join(root, path))
        for path in invalid_paths:
            write_source(os.path.join(root, path), valid=False)
        compiler = Compiler(PythonInterpreter.get())
        yield root, compiler.compile(root, compile_paths)


def test_compile_success():
    # type: () -> None
    with compilation(
        valid_paths=["a.py", "c/c.py"],
        invalid_paths=["b.py", "d/d.py"],
        compile_paths=["a.py", "c/c.py"],
    ) as (root, compiled_relpaths):

        assert 2 == len(compiled_relpaths)

        results = {}
        for compiled in compiled_relpaths:
            compiled_abspath = os.path.join(root, compiled)
            with open(compiled_abspath, "rb") as fp:
                fp.read(4)  # Skip the magic header.
                if sys.version_info[:2] >= (3, 7):
                    # We're in PEP-552 mode: https://www.python.org/dev/peps/pep-0552
                    fp.read(4)  # Skip the invalidation mode bitfield.
                fp.read(4)  # Skip the timestamp.
                if compatibility.PY3:
                    fp.read(4)  # Skip the size.
                code = marshal.load(fp)
            local_symbols = {}  # type: Dict[str, str]
            exec (code, {}, local_symbols)
            results[compiled] = local_symbols

        assert {"basename": "a.py"} == results.pop("a.pyc")
        assert {"basename": "c.py"} == results.pop("c/c.pyc")
        assert 0 == len(results)


def test_compile_failure():
    # type: () -> None
    with pytest.raises(Compiler.Error) as e:
        with compilation(
            valid_paths=["a.py", "c/c.py"],
            invalid_paths=["b.py", "d/d.py"],
            compile_paths=["a.py", "b.py", "c/c.py", "d/d.py"],
        ):
            raise AssertionError("Should not reach here.")

    message = str(e.value)  # type: ignore[unreachable]
    assert "a.py" not in message
    assert "b.py" in message
    assert "c/c.py" not in message
    assert "d/d.py" in message



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import glob
import os.path
import subprocess
import sys
from textwrap import dedent

import pytest

from pex.common import touch
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 7),
    reason="This test needs to run Poetry which requires at least Python 3.7",
)
def test_wheel_file_url_dep(tmpdir):
    # type: (Any) -> None

    constraints = os.path.join(str(tmpdir), "constraints.txt")
    with open(constraints, "w") as fp:
        # The 20.22 release introduces a change that breaks resolution of poetry 1.3.2; so we pin
        # low.
        print("virtualenv<20.22", file=fp)
        # The poetry-plugin-export 1.4.0 release requires poetry>1.5 but poetry 1.3.2 floats the
        # poetry-plugin-export dep; so we pin low.
        print("poetry-plugin-export<1.4", file=fp)

    poetry = os.path.join(str(tmpdir), "poetry.pex")
    run_pex_command(
        args=["poetry==1.3.2", "--constraints", constraints, "-c", "poetry", "-o", poetry]
    ).assert_success()

    corelibrary = os.path.join(str(tmpdir), "corelibrary")
    touch(os.path.join(corelibrary, "README.md"))
    with open(os.path.join(corelibrary, "corelibrary.py"), "w") as fp:
        print("TWO = 2", file=fp)
    subprocess.check_call(args=[poetry, "init", "--no-interaction"], cwd=corelibrary)

    anotherlibrary = os.path.join(str(tmpdir), "anotherlibrary")
    touch(os.path.join(anotherlibrary, "README.md"))
    with open(os.path.join(anotherlibrary, "anotherlibrary.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from corelibrary import TWO


                MEANING_OF_LIFE = TWO * 21
                """
            )
        )
    subprocess.check_call(
        args=[poetry, "init", "--no-interaction", "--dependency", "../corelibrary"],
        cwd=anotherlibrary,
    )

    mylibrary = os.path.join(str(tmpdir), "mylibrary")
    touch(os.path.join(mylibrary, "README.md"))
    with open(os.path.join(mylibrary, "mylibrary.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from anotherlibrary import MEANING_OF_LIFE


                def deep_thought():
                    print(f"MEANING_OF_LIFE = {MEANING_OF_LIFE}")
                """
            )
        )
    subprocess.check_call(
        args=[poetry, "init", "--no-interaction", "--dependency", "../anotherlibrary"],
        cwd=mylibrary,
    )

    subprocess.check_call(args=[poetry, "build", "-f", "wheel"], cwd=mylibrary)
    wheels = glob.glob(os.path.join(mylibrary, "dist", "*.whl"))
    assert len(wheels) == 1
    wheel = wheels[0]

    pex_root = os.path.join(str(tmpdir), "pex_root")
    testing_pex = os.path.join(str(tmpdir), "testing.pex")
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--resolver-version",
            "pip-2020-resolver",
            # N.B.: Modern Pip is needed to handle Poetry relative path deps. Older Pip does
            # building off in a tmp dir and that breaks relative path references.
            "--pip-version",
            "22.3",
            wheel,
            "-e",
            "mylibrary:deep_thought",
            "-o",
            testing_pex,
        ]
    ).assert_success()
    assert (
        "MEANING_OF_LIFE = 42"
        == subprocess.check_output(args=[testing_pex]).decode("utf-8").strip()
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os

from pex.cli.testing import run_pex3
from pex.interpreter import PythonInterpreter
from pex.pep_425 import CompatibilityTags
from pex.pep_508 import MarkerEnvironment
from pex.testing import IntegResults
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Any, Dict, Text


def inspect(
    *args,  # type: str
    **popen_kwargs  # type: str
):
    # type: (...) -> IntegResults
    return run_pex3("interpreter", "inspect", *args, **popen_kwargs)


def assert_inspect(
    *args,  # type: str
    **popen_kwargs  # type: str
):
    # type: (...) -> Text

    result = inspect(*args, **popen_kwargs)
    result.assert_success()
    return result.output


def test_inspect_default(current_interpreter):
    # type: (PythonInterpreter) -> None
    assert current_interpreter.binary == assert_inspect().strip()


def assert_default_verbose_data(
    *args,  # type: str
    **popen_kwargs  # type: str
):
    # type: (...) -> Dict[str, Any]

    data = json.loads(assert_inspect(*args, **popen_kwargs))
    assert PythonInterpreter.get().binary == data.pop("path")
    return cast("Dict[str, Any]", data)


def test_inspect_default_verbose():
    # type: () -> None

    data = assert_default_verbose_data("-v")
    assert data, "There should be additional verbose details beyond just 'path'."


def assert_marker_environment(data):
    # type: (Dict[str, Any]) -> None
    assert PythonInterpreter.get().identity.env_markers == MarkerEnvironment(
        **data.pop("marker_environment")
    )


def test_inspect_default_markers():
    # type: () -> None

    data = assert_default_verbose_data("-m")
    assert_marker_environment(data)
    assert not data


def assert_compatible_tags(data):
    assert PythonInterpreter.get().identity.supported_tags == CompatibilityTags.from_strings(
        data.pop("compatible_tags")
    )


def test_inspect_default_tags():
    # type: () -> None

    data = assert_default_verbose_data("-t")
    assert_compatible_tags(data)
    assert not data


def test_inspect_default_combined():
    # type: () -> None

    data = assert_default_verbose_data("-vmt")
    assert_marker_environment(data)
    assert_compatible_tags(data)
    assert data, (
        "There should be additional verbose details beyond just 'path', 'marker_environment' and "
        "'compatible_tags'."
    )


def test_inspect_all():
    # type: () -> None
    assert [pi.binary for pi in PythonInterpreter.all()] == assert_inspect("--all").splitlines()


def test_inspect_interpreter_selection(
    py27,  # type: PythonInterpreter
    py38,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    assert [py27.binary, py310.binary] == assert_inspect(
        "--python", py27.binary, "--python", py310.binary
    ).splitlines()

    assert [py38.binary, py310.binary] == assert_inspect(
        "--all", "--python-path", os.pathsep.join([os.path.dirname(py38.binary), py310.binary])
    ).splitlines()

    assert [py38.binary] == assert_inspect(
        "--interpreter-constraint",
        "<3.10",
        "--python-path",
        os.pathsep.join([os.path.dirname(py38.binary), py310.binary]),
    ).splitlines()



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from argparse import Action


class HandleBoolAction(Action):
    """An action for bool options that parses `--no-foo` or `--not-foo` as `--foo=False`"""

    def __init__(self, *args, **kwargs):
        kwargs["nargs"] = 0
        super(HandleBoolAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_str=None):
        setattr(namespace, self.dest, not option_str.startswith("--no"))



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import shutil
import subprocess
import sys
from textwrap import dedent

import pytest
from colors import colors

from pex.layout import Layout
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, List


@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
@pytest.mark.parametrize(
    "execution_mode_args",
    [
        pytest.param([], id="UNZIP"),
        pytest.param(["--venv", "--venv-site-packages-copies"], id="VENV (copies)"),
        pytest.param(["--venv", "--no-venv-site-packages-copies"], id="VENV (symlinks)"),
    ],
)
def test_unpack_robustness(
    tmpdir,  # type: Any
    layout,  # type: Layout.Value
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None
    exe = os.path.join(str(tmpdir), "exe.py")
    with open(exe, "w") as fp:
        fp.write(
            dedent(
                """\
                import colors

                print(colors.cyan("Wowbagger hasn't gotten to me yet."))
                """
            )
        )

    pex = os.path.join(str(tmpdir), "pex")
    pex_root = os.path.join(str(tmpdir), "pex_root")
    run_pex_command(
        args=[
            "--runtime-pex-root",
            pex_root,
            "ansicolors==1.1.8",
            "--exe",
            exe,
            "--layout",
            layout.value,
            "-o",
            pex,
        ]
        + execution_mode_args
    ).assert_success()

    def assert_pex_works(pex_path):
        # type: (str) -> None
        assert (
            colors.cyan("Wowbagger hasn't gotten to me yet.")
            == subprocess.check_output(args=[sys.executable, pex_path]).decode("utf-8").strip()
        )

    assert_pex_works(pex)

    elsewhere = os.path.join(str(tmpdir), "elsewhere")
    os.mkdir(elsewhere)
    dest = os.path.join(elsewhere, "other")
    shutil.move(pex, dest)
    assert_pex_works(dest)

    shutil.rmtree(pex_root)
    assert_pex_works(dest)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from textwrap import dedent

import pytest
from colors import yellow

from pex.common import safe_open, touch
from pex.testing import IS_PYPY3, make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterable, List, Text


def create_pex(
    tmpdir,  # type: Any
    include_srcs=True,  # type: bool
    include_deps=True,  # type: bool
    extra_args=(),  # type: Iterable[str]
):
    pex_root = os.path.join(str(tmpdir), "pex_root")
    args = ["--pex-root", pex_root, "--runtime-pex-root", pex_root]

    if include_deps and include_srcs:
        name = "app.pex"
    elif include_deps:
        name = "deps.pex"
    elif include_srcs:
        name = "srcs.pex"
    else:
        name = "pex.file"
    pex_file = os.path.join(str(tmpdir), name)
    args.extend(["-o", pex_file])

    if include_srcs:
        src = os.path.join(str(tmpdir), "src")
        with safe_open(os.path.join(src, "app.py"), "w") as fp:
            fp.write(
                dedent(
                    """\
                    from __future__ import print_function

                    import sys

                    from colors import yellow

                    print(yellow("*** Flashy UI ***"), file=sys.stderr)
                    for entry in sys.path:
                        if not entry.startswith({unzipped_pexes_dir!r}):
                            print(entry)
                    """.format(
                        unzipped_pexes_dir=os.path.join(pex_root, "unzipped_pexes")
                    )
                )
            )
        args.extend(["-D", src, "-m", "app"])

    if include_deps:
        # N.B.: We use this particular version of ansicolors since it contains a top-level module:
        # colors.py. Some of the --venv symlink mode tests below only trigger for the exact case
        # of a distribution with no top level packages and just a top-level modules.
        args.append("ansicolors==1.0.2")

    run_pex_command(args=args + list(extra_args)).assert_success()
    return pex_file


def execute_app(
    pex_file,  # type: str
    **env  # type: str
):
    # type: (...) -> List[Text]

    process = subprocess.Popen(
        args=[pex_file], env=make_env(**env), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = process.communicate()
    stripped_stderr = stderr.decode("utf-8").strip()
    assert 0 == process.returncode, stripped_stderr
    assert yellow("*** Flashy UI ***") in stripped_stderr
    return stdout.decode("utf-8").splitlines()


def test_pex_path_dedup(tmpdir):
    # type: (Any) -> None

    app = create_pex(tmpdir)
    expected_sys_path = execute_app(app)

    deps = create_pex(tmpdir, include_srcs=False)
    assert expected_sys_path == execute_app(app, PEX_PATH=deps)

    srcs = create_pex(tmpdir, include_deps=False)
    assert expected_sys_path == execute_app(app, PEX_PATH=os.pathsep.join((deps, srcs)))


@pytest.mark.parametrize(
    ["execution_mode_args"],
    [
        pytest.param([], id="zipapp"),
        pytest.param(["--venv", "--venv-site-packages-copies"], id="venv (site-packages copies)"),
        pytest.param(
            ["--venv", "--no-venv-site-packages-copies"],
            id="venv (site-packages symlinks)",
        ),
    ],
)
def test_pex_path_collision_non_conflicting(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    app = create_pex(tmpdir, extra_args=execution_mode_args)
    execute_app(app)

    # Test a non-conflicting duplicate dep collision.
    deps = create_pex(tmpdir, include_srcs=False)
    execute_app(app, PEX_PATH=deps)

    # Test a non-conflicting duplicate src collision.
    srcs = create_pex(tmpdir, include_deps=False)
    execute_app(app, PEX_PATH=os.pathsep.join((deps, srcs)))


def test_pex_path_collision_conflicting(tmpdir):
    # Test a conflicting duplicate dep top-level module collision.

    conflicting_colors = os.path.join(str(tmpdir), "conflicting_colors")
    touch(os.path.join(conflicting_colors, "colors.py"))
    with safe_open(os.path.join(conflicting_colors, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from setuptools import setup

                setup(
                    name="conflicting-colors",
                    version="0.0.1",
                    py_modules=["colors"]
                )
                """
            )
        )
    alternate_colors = create_pex(
        tmpdir, include_srcs=False, include_deps=False, extra_args=[conflicting_colors]
    )
    env = make_env(PEX_PATH=alternate_colors)

    # A zipapp with a collision should always work, 1st on PEX_PATH wins.
    app = create_pex(tmpdir)
    execute_app(app, **env)

    def assert_venv_collision(*extra_args):
        # type: (*str) -> None

        app = create_pex(tmpdir, extra_args=["--venv"] + list(extra_args))
        process = subprocess.Popen(
            args=[app], env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        _, stderr = process.communicate()
        decoded_stderr = stderr.decode("utf-8")

        # The --venv mode should warn about collisions but succeed.
        assert 0 == process.returncode, decoded_stderr
        assert "PEXWarning: Encountered collision populating " in decoded_stderr
        assert "site-packages/colors.py was provided by:" in decoded_stderr
        assert "sha1:17772af8295ffb7f4d6c3353665b5c542be332a2 -> " in decoded_stderr
        assert "sha1:da39a3ee5e6b4b0d3255bfef95601890afd80709 -> " in decoded_stderr

    assert_venv_collision("--venv-site-packages-copies")
    assert_venv_collision("--no-venv-site-packages-copies")



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import errno
import os
from contextlib import contextmanager

import pytest

from pex.atomic_directory import AtomicDirectory, FileLockStyle, _is_bsd_lock, atomic_directory
from pex.common import temporary_dir, touch
from pex.testing import environment_as
from pex.typing import TYPE_CHECKING

try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore[no-redef,import]

if TYPE_CHECKING:
    from typing import Iterator, Optional, Type


def test_is_bsd_lock():
    # type: () -> None

    assert not _is_bsd_lock(
        lock_style=None
    ), "Expected the default lock style to be POSIX for maximum compatibility."
    assert not _is_bsd_lock(lock_style=FileLockStyle.POSIX)
    assert _is_bsd_lock(lock_style=FileLockStyle.BSD)

    # The hard-coded default is already POSIX, so setting the env var default changes nothing.
    with environment_as(_PEX_FILE_LOCK_STYLE="posix"):
        assert not _is_bsd_lock(lock_style=None)
        assert not _is_bsd_lock(lock_style=FileLockStyle.POSIX)
        assert _is_bsd_lock(lock_style=FileLockStyle.BSD)

    with environment_as(_PEX_FILE_LOCK_STYLE="bsd"):
        assert _is_bsd_lock(
            lock_style=None
        ), "Expected the default lock style to be taken from the environment when defined."
        assert not _is_bsd_lock(lock_style=FileLockStyle.POSIX)
        assert _is_bsd_lock(lock_style=FileLockStyle.BSD)


@contextmanager
def maybe_raises(exception=None):
    # type: (Optional[Type[Exception]]) -> Iterator[None]
    @contextmanager
    def noop():
        yield

    context = noop() if exception is None else pytest.raises(exception)
    with context:
        yield


def atomic_directory_finalize_test(errno, expect_raises=None):
    # type: (int, Optional[Type[Exception]]) -> None
    with mock.patch("os.rename", spec_set=True, autospec=True) as mock_rename:
        mock_rename.side_effect = OSError(errno, os.strerror(errno))
        with maybe_raises(expect_raises):
            AtomicDirectory("to.dir").finalize()


def test_atomic_directory_finalize_eexist():
    # type: () -> None
    atomic_directory_finalize_test(errno.EEXIST)


def test_atomic_directory_finalize_enotempty():
    # type: () -> None
    atomic_directory_finalize_test(errno.ENOTEMPTY)


def test_atomic_directory_finalize_eperm():
    # type: () -> None
    atomic_directory_finalize_test(errno.EPERM, expect_raises=OSError)


def test_atomic_directory_empty_workdir_finalize():
    # type: () -> None
    with temporary_dir() as sandbox:
        target_dir = os.path.join(sandbox, "target_dir")
        assert not os.path.exists(target_dir)

        with atomic_directory(target_dir) as atomic_dir:
            assert not atomic_dir.is_finalized()
            assert target_dir == atomic_dir.target_dir
            assert os.path.exists(atomic_dir.work_dir)
            assert os.path.isdir(atomic_dir.work_dir)
            assert [] == os.listdir(atomic_dir.work_dir)

            touch(os.path.join(atomic_dir.work_dir, "created"))

            assert not os.path.exists(target_dir)

        assert not os.path.exists(atomic_dir.work_dir), "The work_dir should always be cleaned up."
        assert os.path.exists(os.path.join(target_dir, "created"))


def test_atomic_directory_empty_workdir_failure():
    # type: () -> None
    class SimulatedRuntimeError(RuntimeError):
        pass

    with temporary_dir() as sandbox:
        target_dir = os.path.join(sandbox, "target_dir")
        with pytest.raises(SimulatedRuntimeError):
            with atomic_directory(target_dir) as atomic_dir:
                assert not atomic_dir.is_finalized()
                touch(os.path.join(atomic_dir.work_dir, "created"))
                raise SimulatedRuntimeError()

        assert not os.path.exists(  # type: ignore[unreachable]
            atomic_dir.work_dir
        ), "The work_dir should always be cleaned up."
        assert not os.path.exists(target_dir), (
            "When the context raises the work_dir it was given should not be moved to the "
            "target_dir."
        )


def test_atomic_directory_empty_workdir_finalized():
    # type: () -> None
    with temporary_dir() as target_dir:
        with atomic_directory(target_dir) as work_dir:
            assert (
                work_dir.is_finalized()
            ), "When the target_dir exists no work_dir should be created."


def test_atomic_directory_locked_mode():
    # type: () -> None

    assert AtomicDirectory("unlocked").work_dir != AtomicDirectory("unlocked").work_dir
    assert (
        AtomicDirectory("locked", locked=True).work_dir
        == AtomicDirectory("locked", locked=True).work_dir
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import weakref
from collections import defaultdict
from functools import total_ordering

from _weakref import ReferenceType

from pex.typing import TYPE_CHECKING, Generic, cast

if TYPE_CHECKING:
    from typing import Any, DefaultDict, List, Optional, Tuple, Type, TypeVar

    _V = TypeVar("_V", bound="Enum.Value")


class Enum(Generic["_V"]):
    @total_ordering
    class Value(object):
        _values_by_type = defaultdict(
            list
        )  # type: DefaultDict[Type[Enum.Value], List[ReferenceType[Enum.Value]]]

        @classmethod
        def _iter_values(cls):
            for ref in cls._values_by_type[cls]:
                value = ref()
                if value:
                    yield value

        def __init__(self, value):
            # type: (str) -> None
            values = Enum.Value._values_by_type[type(self)]
            self.value = value
            self.ordinal = len(values)
            values.append(weakref.ref(self))

        def __str__(self):
            # type: () -> str
            return str(self.value)

        def __repr__(self):
            # type: () -> str
            return repr(self.value)

        def __eq__(self, other):
            # type: (Any) -> bool
            return self is other

        def __hash__(self):
            # type: () -> int
            return hash(self.value)

        @classmethod
        def _create_type_error(cls, other):
            # type: (Any) -> TypeError
            return TypeError(
                "Can only compare values of type {value_type} amongst themselves; given "
                "{other!r} of type {other_type}.".format(
                    value_type=qualified_name(cls),
                    other=other,
                    other_type=qualified_name(other),
                )
            )

        def __lt__(self, other):
            # type: (Any) -> bool
            if type(self) != type(other):
                raise self._create_type_error(other)
            return self.ordinal < cast(Enum.Value, other).ordinal

        def __le__(self, other):
            # type: (Any) -> bool
            if type(self) != type(other):
                raise self._create_type_error(other)
            return self is other or self < other

    _values = None  # type: Optional[Tuple[_V, ...]]

    @classmethod
    def values(cls):
        # type: (Type[Enum[_V]]) -> Tuple[_V, ...]
        if cls._values is None:
            cls._values = tuple(cls.Value._iter_values())
        return cls._values

    @classmethod
    def for_value(
        cls,  # type: Type[Enum[_V]]
        value,  # type: str
    ):
        # type: (...) -> _V
        for v in cls.values():
            if v.value == value:
                return v
        raise ValueError(
            "{!r} of type {} must be one of {}".format(
                value, type(value), ", ".join(map(repr, cls.values()))
            )
        )


def qualified_name(item):
    # type: (Any) -> str
    """Attempt to produce the fully qualified name for an item.

    If the item is a type, method, property or function, its fully qualified name is returned as
    best as can be determined. Otherwise, the fully qualified name of the type of the given item is
    returned.

    :param item: The item to identify.
    :return: The fully qualified name of the given item.
    """
    if isinstance(item, property):
        item = item.fget
    if not hasattr(item, "__name__"):
        item = type(item)
    return "{module}.{type}".format(
        module=getattr(item, "__module__", "<unknown module>"),
        # There is no __qualname__ in Python 2.7; so we do the best we can.
        type=getattr(item, "__qualname__", item.__name__),
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess

import pytest

from pex.common import temporary_dir
from pex.testing import IS_PYPY, run_pex_command


@pytest.mark.skipif(
    IS_PYPY,
    reason="On PyPy this causes this error: Failed to execute PEX file. Needed "
    "manylinux2014_x86_64-pp-272-pypy_41 compatible dependencies for 1: "
    "cryptography==2.5 But this pex only contains "
    "cryptography-2.5-pp27-pypy_41-linux_x86_64.whl. "
    "Temporarily skipping the test on PyPy allows us to get tests passing "
    "again, until we can address this.",
)
def test_devendoring_required():
    # type: () -> None
    # The cryptography distribution does not have a whl released for python3 on linux at version 2.5.
    # As a result, we're forced to build it under python3 and, prior to the fix for
    # https://github.com/pantsbuild/pex/issues/661, this would fail using the vendored setuptools
    # inside pex.
    with temporary_dir() as td:
        cryptography_pex = os.path.join(td, "cryptography.pex")
        res = run_pex_command(["cryptography==2.5", "-o", cryptography_pex])
        res.assert_success()

        subprocess.check_call([cryptography_pex, "-c", "import cryptography"])



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import sqlite3
from contextlib import closing
from itertools import repeat
from multiprocessing.pool import ThreadPool

from pex import pex_warnings
from pex.compatibility import cpu_count
from pex.fetcher import URLFetcher
from pex.resolve.pep_691.api import Client
from pex.resolve.pep_691.model import Endpoint, Project
from pex.resolve.resolved_requirement import Fingerprint, PartialArtifact
from pex.resolve.resolvers import MAX_PARALLEL_DOWNLOADS
from pex.result import Error, catch
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Iterable, Iterator, List, Optional, Sequence, Set, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class _FingerprintedURL(object):
    url = attr.ib()  # type: str
    fingerprint = attr.ib()  # type: Fingerprint


@attr.s(frozen=True)
class FingerprintService(object):
    @classmethod
    def create(
        cls,
        url_fetcher,  # type: URLFetcher
        max_parallel_jobs=None,  # type: Optional[int]
    ):
        return cls(api=Client(url_fetcher=url_fetcher), max_parallel_jobs=max_parallel_jobs)

    _api = attr.ib(factory=Client)  # type: Client
    _path = attr.ib(factory=lambda: os.path.join(ENV.PEX_ROOT, "fingerprints.db"))  # type: str
    _max_parallel_jobs = attr.ib(default=None)  # type: Optional[int]

    @property
    def accept(self):
        # type: () -> Tuple[str, ...]
        return self._api.ACCEPT

    def _iter_cached(self, urls_to_fingerprint):
        # type: (Iterable[str]) -> Iterator[_FingerprintedURL]

        urls = sorted(urls_to_fingerprint)
        with TRACER.timed("Searching for {count} fingerprints in database".format(count=len(urls))):
            with sqlite3.connect(self._path) as conn:
                conn.executescript(
                    """
                    CREATE TABLE IF NOT EXISTS hashes (
                        url TEXT PRIMARY KEY ASC,
                        algorithm TEXT NOT NULL,
                        hash TEXT NOT NULL
                    ) WITHOUT ROWID;
                    """
                ).close()
                # N.B.: Maximum parameter count is 999 in pre-2020 versions of SQLite 3; so we limit
                # to an even lower chunk size to be safe: https://www.sqlite.org/limits.html
                chunk_size = 100
                for index in range(0, len(urls), chunk_size):
                    chunk = urls[index : index + chunk_size]
                    with closing(
                        conn.execute(
                            "SELECT url, algorithm, hash FROM hashes WHERE url IN ({})".format(
                                ", ".join(repeat("?", len(chunk)))
                            ),
                            tuple(chunk),
                        )
                    ) as cursor:
                        for url, algorithm, hash_ in cursor:
                            yield _FingerprintedURL(
                                url=url, fingerprint=Fingerprint(algorithm=algorithm, hash=hash_)
                            )

    def _cache(self, fingerprinted_urls):
        # type: (Sequence[_FingerprintedURL]) -> None

        with TRACER.timed(
            "Caching {count} fingerprints in database".format(count=len(fingerprinted_urls))
        ):
            with sqlite3.connect(self._path) as conn:
                conn.executemany(
                    "INSERT OR REPLACE INTO hashes (url, algorithm, hash) VALUES (?, ?, ?)",
                    tuple(
                        (
                            fingerprinted_url.url,
                            fingerprinted_url.fingerprint.algorithm,
                            fingerprinted_url.fingerprint.hash,
                        )
                        for fingerprinted_url in fingerprinted_urls
                    ),
                ).close()

    def _safe_request(self, endpoint):
        # type: (Endpoint) -> Union[Project, Error]
        return catch(self._api.request, endpoint)

    @staticmethod
    def _warn_database_error(
        error,  # type: sqlite3.DatabaseError
        message,  # type: str
    ):
        pex_warnings.warn(
            "{message}: {error}\n"
            "\n"
            "If you encounter this error frequently, Pex may need to adjust how it uses SQLite for "
            "its cache of artifact fingerprints obtained from PEP-691 indexes. Please consider "
            "reporting the problem by filing an issue at "
            "https://github.com/pantsbuild/pex/issues/new.".format(message=message, error=error)
        )

    def fingerprint(
        self,
        endpoints,  # type: Set[Endpoint]
        artifacts,  # type: Iterable[PartialArtifact]
    ):
        # type: (...) -> Iterator[PartialArtifact]
        """Attempts to fingerprint all artifacts that are missing a fingerprint.

        Fingerprints are obtained via the PEP-691 JSON API and are not verified.

        :param artifacts: The artifacts to fill in missing fingerprints for.
        :return: An iterator over all the artifacts given, where some returned artifacts may have
                 previously missing fingerprints filled in (but not verified).
        """
        # N.B.: In the absence of having an advertised fingerprint for an artifact, Pex will
        # download the artifact and hash it at higher layers; so simply warning about the error is
        # enough; we need not fail the whole endeavor. The partial artifact just remains partial
        # and becomes more expensive to lock. The warning is likely un-actionable by the user but
        # also rare; so it can be reported to maintainers in case the 5s default sqlite3 timeout is
        # being hit or something similar that might be mitigated by adding more code here.

        artifacts_to_fingerprint = {}
        for artifact in artifacts:
            if artifact.fingerprint:
                yield artifact
            else:
                artifacts_to_fingerprint[artifact.url.normalized_url] = artifact

        if not artifacts_to_fingerprint:
            return

        try:
            cached = 0
            for fingerprinted_url in self._iter_cached(artifacts_to_fingerprint):
                yield attr.evolve(
                    artifacts_to_fingerprint[fingerprinted_url.url],
                    fingerprint=fingerprinted_url.fingerprint,
                )
                artifacts_to_fingerprint.pop(fingerprinted_url.url)
                cached += 1
            TRACER.log("Found {count} fingerprints cached in database.".format(count=cached))
        except sqlite3.DatabaseError as e:
            self._warn_database_error(
                error=e,
                message=(
                    "Failed to read fingerprints from the cache, continuing to fetch them via the "
                    "PEP-691 JSON API instead"
                ),
            )

        if not artifacts_to_fingerprint:
            return

        max_threads = min(
            len(endpoints) or 1,
            min(MAX_PARALLEL_DOWNLOADS, 4 * (self._max_parallel_jobs or cpu_count() or 1)),
        )
        with TRACER.timed(
            "Making {api_count} PEP-691 JSON API requests across {thread_count} threads to "
            "fingerprint {artifact_count} artifacts".format(
                api_count=len(endpoints),
                thread_count=max_threads,
                artifact_count=len(artifacts_to_fingerprint),
            )
        ):
            pool = ThreadPool(processes=max_threads)
            try:
                api_results = pool.map(self._safe_request, endpoints)
            finally:
                pool.close()
                pool.join()

        fingerprinted_urls = []  # type: List[_FingerprintedURL]
        for result in api_results:
            if isinstance(result, Error):
                pex_warnings.warn(
                    "Failed to fetch project metadata, continuing: {error}".format(error=result)
                )
                continue

            for file in result.files:
                fingerprinted_artifact = artifacts_to_fingerprint.pop(file.url, None)
                if not fingerprinted_artifact:
                    continue

                maybe_fingerprint = file.select_fingerprint()
                yield attr.evolve(fingerprinted_artifact, fingerprint=maybe_fingerprint)
                if maybe_fingerprint:
                    fingerprinted_urls.append(
                        _FingerprintedURL(url=file.url, fingerprint=maybe_fingerprint)
                    )

        # The remaining artifacts have no fingerprint and no endpoint to fetch the data from; so we
        # just return them as-is.
        for artifact in artifacts_to_fingerprint.values():
            yield artifact

        if not fingerprinted_urls:
            return

        try:
            self._cache(fingerprinted_urls)
        except sqlite3.DatabaseError as e:
            self._warn_database_error(
                error=e,
                message="Failed to cache fingerprints for {count} URLs, continuing".format(
                    count=len(fingerprinted_urls)
                ),
            )



# Copyright 2017 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

# A library of functions for filtering Python interpreters based on compatibility constraints

from __future__ import absolute_import

import itertools

from pex.compatibility import indent
from pex.dist_metadata import ProjectNameAndVersion, Requirement, RequirementParseError
from pex.enum import Enum
from pex.interpreter import PythonInterpreter
from pex.orderedset import OrderedSet
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Iterator, Optional, Tuple

    import attr  # vendor:skip

    from pex.interpreter import InterpreterIdentificationError
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class InterpreterConstraint(object):
    @classmethod
    def parse(
        cls,
        constraint,  # type: str
        default_interpreter=None,  # type: Optional[str]
    ):
        # type: (...) -> InterpreterConstraint
        try:
            requirement = Requirement.parse(constraint)
            return cls(specifier=requirement.specifier, name=requirement.name)
        except RequirementParseError:
            if default_interpreter is not None:
                return cls.parse(
                    constraint="{interpreter}{specifier}".format(
                        interpreter=default_interpreter, specifier=constraint
                    ),
                    default_interpreter=None,
                )
            try:
                return cls(specifier=SpecifierSet(constraint))
            except RequirementParseError as e:
                raise ValueError(
                    "Unparseable interpreter constraint {constraint}: {err}".format(
                        constraint=constraint, err=e
                    )
                )

    @classmethod
    def matches(
        cls,
        constraint,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        return (interpreter or PythonInterpreter.get()) in cls.parse(constraint)

    @classmethod
    def exact_version(cls, interpreter=None):
        # type: (Optional[PythonInterpreter]) -> InterpreterConstraint
        python_identity = (interpreter or PythonInterpreter.get()).identity
        return cls.parse(
            "{interpreter}=={version}".format(
                interpreter=python_identity.interpreter, version=python_identity.version_str
            )
        )

    specifier = attr.ib()  # type: SpecifierSet
    name = attr.ib(default=None)  # type: Optional[str]

    def iter_matching(self, paths=None):
        # type: (Optional[Iterable[str]]) -> Iterator[PythonInterpreter]
        for interp in PythonInterpreter.iter(paths=paths):
            if interp in self:
                yield interp

    @property
    def requires_python(self):
        # type: () -> SpecifierSet
        return self.specifier

    def __str__(self):
        # type: () -> str
        return "{name}{specifier}".format(name=self.name or "", specifier=self.specifier)

    def __contains__(self, interpreter):
        # type: (PythonInterpreter) -> bool
        python_identity = interpreter.identity
        if self.name and self.name != python_identity.interpreter:
            return False
        return python_identity.version_str in self.specifier


@attr.s(frozen=True)
class InterpreterConstraints(object):
    @classmethod
    def parse(cls, *constraints):
        # type: (str) -> InterpreterConstraints
        return cls(
            constraints=tuple(
                InterpreterConstraint.parse(constraint) for constraint in OrderedSet(constraints)
            )
        )

    constraints = attr.ib(default=())  # type: Tuple[InterpreterConstraint, ...]

    def merged(self, other):
        # type: (InterpreterConstraints) -> InterpreterConstraints
        constraints = OrderedSet(self.constraints)
        constraints.update(other.constraints)
        return InterpreterConstraints(constraints=tuple(constraints))

    def __str__(self):
        # type: () -> str
        return " or ".join(str(constraint) for constraint in self.constraints)

    def __contains__(self, interpreter):
        if not self.constraints:
            return True
        return any(interpreter in constraint for constraint in self.constraints)

    def __iter__(self):
        # type: () -> Iterator[InterpreterConstraint]
        return iter(self.constraints)

    def __bool__(self):
        # type: () -> bool
        return bool(self.constraints)

    def __len__(self):
        # type: () -> int
        return len(self.constraints)

    def __getitem__(self, item):
        # type: (int) -> InterpreterConstraint
        return self.constraints[item]


class UnsatisfiableInterpreterConstraintsError(Exception):
    """Indicates interpreter constraints could not be satisfied."""

    def __init__(
        self,
        constraints,  # type: Iterable[str]
        candidates,  # type: Iterable[PythonInterpreter]
        failures,  # type: Iterable[InterpreterIdentificationError]
        preamble=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        """
        :param constraints: The constraints that could not be satisfied.
        :param candidates: The python interpreters that were compared against the constraints.
        :param failures: Descriptions of the python interpreters that were unidentifiable.
        :param preamble: An optional preamble for the exception message.
        """
        self.constraints = tuple(constraints)
        self.candidates = tuple(candidates)
        self.failures = tuple(failures)
        super(UnsatisfiableInterpreterConstraintsError, self).__init__(
            self.create_message(preamble=preamble)
        )

    def with_preamble(self, preamble):
        # type: (str) -> UnsatisfiableInterpreterConstraintsError
        return UnsatisfiableInterpreterConstraintsError(
            self.constraints, self.candidates, self.failures, preamble=preamble
        )

    def create_message(self, preamble=None):
        # type: (Optional[str]) -> str
        """Create a message describing  failure to find matching interpreters with an optional
        preamble.

        :param preamble: An optional preamble to the message that will be displayed above it
                             separated by an empty blank line.
        :return: A descriptive message usable for display to an end user.
        """
        preamble = "{}\n\n".format(preamble) if preamble else ""

        failures_message = ""
        if self.failures:
            seen = set()
            broken_interpreters = []
            for python, error in self.failures:
                canonical_python = PythonInterpreter.canonicalize_path(python)
                if canonical_python not in seen:
                    broken_interpreters.append((canonical_python, error))
                    seen.add(canonical_python)

            failures_message = (
                "{}\n"
                "\n"
                "(See https://github.com/pantsbuild/pex/issues/1027 for a list of known breaks and "
                "workarounds.)"
            ).format(
                "\n".join(
                    "{index}.) {binary}:\n{error}".format(index=i, binary=python, error=error)
                    for i, (python, error) in enumerate(broken_interpreters, start=1)
                )
            )

        if not self.candidates:
            if failures_message:
                return (
                    "{preamble}"
                    "Interpreters were found but they all appear to be broken:\n"
                    "{failures}"
                ).format(preamble=preamble, failures=failures_message)
            return "{}No interpreters could be found on the system.".format(preamble)

        binary_column_width = max(len(candidate.binary) for candidate in self.candidates)
        interpreters_format = "{{index}}.) {{binary: >{}}} {{requirement}}".format(
            binary_column_width
        )

        qualifier = ""
        if failures_message:
            failures_message = "Skipped the following broken interpreters:\n{}".format(
                failures_message
            )
            qualifier = "working "

        constraints_message = ""
        if self.constraints:
            constraints_message = (
                "No {qualifier}interpreter compatible with the requested constraints was found:\n"
                "\n{constraints}"
            ).format(
                qualifier=qualifier,
                constraints="\n\n".join(
                    indent(constraint, "  ") for constraint in self.constraints
                ),
            )

        problems = "\n\n".join(msg for msg in (failures_message, constraints_message) if msg)
        if problems:
            problems = "\n\n{}".format(problems)

        return (
            "{preamble}"
            "Examined the following {qualifier}interpreters:\n"
            "{interpreters}"
            "{problems}"
        ).format(
            preamble=preamble,
            qualifier=qualifier,
            interpreters="\n".join(
                interpreters_format.format(
                    index=i,
                    binary=candidate.binary,
                    requirement=InterpreterConstraint.exact_version(candidate),
                )
                for i, candidate in enumerate(self.candidates, start=1)
            ),
            problems=problems,
        )


class Lifecycle(Enum["Lifecycle.Value"]):
    class Value(Enum.Value):
        pass

    DEV = Value("dev")
    STABLE = Value("stable")
    EOL = Value("eol")


# This value is based off of:
# 1. Past releases: https://www.python.org/downloads/ where the max patch level was achieved by
#    2.7.18.
# 2. The 3.9+ annual release cycle formalization: https://www.python.org/dev/peps/pep-0602/ where
#    the last bugfix release will be at a patch level of ~10 and then 3.5 years of security fixes
#    as needed before going to EOL at the 5-year mark.
DEFAULT_MAX_PATCH = 30


@attr.s(frozen=True)
class PythonVersion(object):
    lifecycle = attr.ib()  # type: Lifecycle.Value
    major = attr.ib()  # type: int
    minor = attr.ib()  # type: int
    patch = attr.ib()  # type: int

    def iter_compatible_versions(
        self,
        specifier_sets,  # type: Iterable[SpecifierSet]
        max_patch=DEFAULT_MAX_PATCH,  # type: int
    ):
        # type: (...) -> Iterator[Tuple[int, int, int]]
        last_patch = self.patch if self.lifecycle == Lifecycle.EOL else max_patch
        for patch in range(last_patch + 1):
            version = (self.major, self.minor, patch)
            version_string = ".".join(map(str, version))
            if not specifier_sets:
                yield version
            else:
                for specifier_set in specifier_sets:
                    if version_string in specifier_set:
                        yield version
                        break


# TODO(John Sirois): Integrate a `pyenv install -l` based lint / generate script for CI / local
# use that emits the current max patch for these versions so we automatically stay up to date
# mod dormancy in the project.

COMPATIBLE_PYTHON_VERSIONS = (
    PythonVersion(Lifecycle.EOL, 2, 7, 18),
    # N.B.: Pex does not support the missing 3.x versions here.
    PythonVersion(Lifecycle.EOL, 3, 5, 10),
    PythonVersion(Lifecycle.EOL, 3, 6, 15),
    PythonVersion(Lifecycle.STABLE, 3, 7, 15),
    PythonVersion(Lifecycle.STABLE, 3, 8, 15),
    PythonVersion(Lifecycle.STABLE, 3, 9, 15),
    PythonVersion(Lifecycle.STABLE, 3, 10, 8),
    PythonVersion(Lifecycle.STABLE, 3, 11, 0),
    PythonVersion(Lifecycle.DEV, 3, 12, 0),
)


def iter_compatible_versions(
    requires_python,  # type: Iterable[str]
    max_patch=DEFAULT_MAX_PATCH,  # type: int
):
    # type: (...) -> Iterator[Tuple[int, int, int]]

    specifier_sets = OrderedSet(SpecifierSet(req) for req in requires_python)
    return itertools.chain.from_iterable(
        python_version.iter_compatible_versions(specifier_sets, max_patch=max_patch)
        for python_version in COMPATIBLE_PYTHON_VERSIONS
    )
